/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *  Copyright (C) 2008 Andrew Burton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __GUI_PREFS_H__
#define __GUI_PREFS_H__

#include <gtk/gtk.h>

G_BEGIN_DECLS

/* Preferences object */
#define PCSX_TYPE_GUI_PREFS            (pcsx_gui_prefs_get_type ())
#define PCSX_GUI_PREFS(o)            (G_TYPE_CHECK_INSTANCE_CAST((o), PCSX_TYPE_GUI_PREFS, PcsxGuiPrefs))
#define PCSX_GUI_PREFS_CLASS(k)    (G_TYPE_CHECK_CLASS_CAST ((k), PCSX_TYPE_GUI_PREFS, PcsxGuiPrefsClass))
#define PCSX_IS_GUI_PREFS(o)         (G_TYPE_CHECK_INSTANCE_TYPE ((o), PCSX_TYPE_GUI_PREFS))
#define PCSX_IS_GUI_PREFS_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), PCSX_TYPE_GUI_PREFS))
#define PCSX_GUI_PREFS_GET_CLASS(o)  (G_TYPE_INSTANCE_GET_CLASS ((o), PCSX_TYPE_GUI_PREFS, PcsxGuiPrefsClass))

typedef struct _PcsxGuiPrefs PcsxGuiPrefs;
typedef struct _PcsxGuiPrefsClass PcsxGuiPrefsClass;
typedef struct _PcsxGuiPrefsPrivate PcsxGuiPrefsPrivate;

/* The GTK widget type, first property in the Glade widget name */
typedef enum
{
	PCSX_PROPERTY_OBJECT_TYPE_TOGGLE,		/* GtkToggleButton */
	PCSX_PROPERTY_OBJECT_TYPE_CONF_BUTTON,  /* GtkButton, used for plugin preferences */
	PCSX_PROPERTY_OBJECT_TYPE_ABOUT_BUTTON, /* GtkButton, used for plugin about */
	PCSX_PROPERTY_OBJECT_TYPE_COMBO,		/* GtkCombo, for Plugins */
	PCSX_PROPERTY_OBJECT_TYPE_BIOSCOMBO,	/* GtkCombo, for BIOS */
	PCSX_PROPERTY_OBJECT_TYPE_FOLDER		/* GtkFileChooser, for plugins and BIOS path */
} PcsxPropertyObjectType;

struct _PcsxGuiPrefs {
	GObject parent;
	
	PcsxGuiPrefsPrivate *priv;
	/* define public instance variables here */
};

struct _PcsxGuiPrefsClass {
	GObjectClass parent;
	/* define vtable methods and signals here */
};

/* Preferences */
enum
{
	PROP_0,

	PROP_GUI_PREFS_BIOS,
	PROP_GUI_PREFS_LOADCDBIOS,
	PROP_GUI_PREFS_PLUGIN_GPU,
	PROP_GUI_PREFS_PLUGIN_SPU,
	PROP_GUI_PREFS_PLUGIN_CDR,
	PROP_GUI_PREFS_PLUGIN_PAD1,
	PROP_GUI_PREFS_PLUGIN_PAD2,

	PROP_GUI_PREFS_DIR_BIOS,
	PROP_GUI_PREFS_DIR_PLUGINS,

	PROP_GUI_PREFS_CPU_XA,
	PROP_GUI_PREFS_CPU_SIO,
	PROP_GUI_PREFS_CPU_MDEC,
	PROP_GUI_PREFS_CPU_PSXAUTO,
	PROP_GUI_PREFS_CPU_PSXTYPE,
	PROP_GUI_PREFS_CPU_CDDA,
	PROP_GUI_PREFS_CPU_CPU,
	PROP_GUI_PREFS_CPU_DBG,
	PROP_GUI_PREFS_CPU_PSXOUT,
	PROP_GUI_PREFS_CPU_SPUIRQ,
	PROP_GUI_PREFS_CPU_RCNTFIX,
	PROP_GUI_PREFS_CPU_VSYNCWA,

	PROP_GUI_PREFS_MEMCARD1,
	PROP_GUI_PREFS_MEMCARD2,

	PROP_GUI_PREFS_CURRENTFILE,
	
	NUM_GUI_PREFS_PROPERTIES
};

GType pcsx_gui_prefs_get_type (void);
PcsxGuiPrefs* pcsx_gui_prefs_new (void);

/* Temporarily accessible from outside */
PcsxGuiPrefs *pcsx_gui_prefs;


G_END_DECLS

#endif /* __GUI_PREFS_H__ */