/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
  
#include <string.h>
#include <dirent.h>
#include <dlfcn.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gio/gio.h>
#include <regex.h>

#include "Gtk2Gui.h"
#include "Linux.h"
#include "configure-dlg.h"
#include "memcard-dlg.h"
#include "pcsx-main-win.h"

#include "hdebug.h"
/* FIXME */
#include "../libpcsxcore/plugins.h"
#include "../libpcsxcore/sio.h"

extern int UseGui;
long LoadCdBios;

// Helper Functions
#ifdef REDUNDANT_CODE
void UpdatePluginsBIOS();
void UpdatePluginsBIOS_UpdateGUI(GladeXML *xml);
void ScanPlugins(gchar*);
void FindNetPlugin(GladeXML *xml);
#endif /* REDUNDANT_CODE */

#ifdef NETWORKING_ENABLED
void OnConf_Net();
void OnNet_Conf(GtkWidget *widget, gpointer user_data);
void OnNet_About(GtkWidget *widget, gpointer user_data);
#endif /* NETWORKING_ENABLED */

#ifdef REDUNDANT_CODE
static void on_file_run_cd (GtkWidget *widget, gpointer user_data);
void OnBiosPath_Changed(GtkWidget *wdg, gpointer data);
void OnConf_Clicked (GtkDialog *dialog, gint arg1, gpointer user_data);
void OnPluginPath_Changed(GtkWidget *wdg, gpointer data);
void OnConfConf_Pad1About(GtkWidget *widget, gpointer user_data);
void OnConfConf_Pad2About(GtkWidget *widget, gpointer user_data);
static void on_configure_plugin (GtkWidget *widget, gpointer user_data);
static void on_states_load (GtkWidget *widget, gpointer user_data);
static void on_states_save (GtkWidget *widget, gpointer user_data);
static void on_states_save_other ();
#endif /* REDUNDANT_CODE */

static void destroy_main_window ();

void SysErrorMessage (gchar *primary, gchar *secondary);
void SysInfoMessage (gchar *primary, gchar *secondary);

#ifdef REDUNDANT_CODE
void on_about_plugin (GtkWidget *widget, gpointer user_data);
static int match(const char* string, char* pattern);
#endif /* REDUNDANT_CODE */

PcsxMainWin *Window;
#ifdef REDUNDANT_CODE
GtkWidget *ConfDlg;
GtkWidget *controlwidget;

#define MAX_SLOTS 5

#define FindComboText(combo,list,conf) \
	if (strlen(conf) > 0) { \
		int i; \
		for (i=2;i<255;i+=2) { \
			if (!strcmp(conf, list[i-2])) { \
				gtk_combo_box_set_active (GTK_COMBO_BOX (combo), i/2-1); \
				break; \
			} \
		} \
	}

#define GetComboText(combo,list,conf) \
	{ \
		int row; \
		row = gtk_combo_box_get_active (GTK_COMBO_BOX (combo)); \
		strcpy(conf, (char *) list[row*2]); \
	}
 
/* TODO - If MAX_SLOTS changes, need to find a way to automatically set all positions */
int Slots[MAX_SLOTS] = { -1, -1, -1, -1, -1 };

void ResetMenuSlots(GladeXML *xml) {
	GtkWidget *widget;
	gchar *str;
	int i;

	for (i = 0; i < MAX_SLOTS; i++) {
		str = g_strdup_printf("GtkMenuItem_LoadSlot%d", i+1);
		widget = glade_xml_get_widget (xml, str);
		g_free (str);
		
		if (Slots[i] == -1) 
			gtk_widget_set_sensitive(widget, FALSE);
		else
			gtk_widget_set_sensitive(widget, TRUE);
	}
}

void UpdateMenuSlots() {
	gchar *str;
	int i;

	for (i = 0; i < MAX_SLOTS; i++) {
		str = get_state_filename (i);
		Slots[i] = CheckState(str);
	}
	
	g_free (str);
}

int match(const char *string, char *pattern)
{
    int    status;
    regex_t    re;

    if (regcomp(&re, pattern, REG_EXTENDED|REG_NOSUB) != 0) {
        return(0);      /* report error */
    }
    status = regexec(&re, string, (size_t) 0, NULL, 0);
    regfree(&re);
    if (status != 0) {
        return(0);      /* report error */
    }
    return(1);
}
#endif /* REDUNDANT_CODE */

/* Builds the filename based on the currently loaded CD image's
   ID and Label. These values are set when the CDR plugin is loaded */
gchar* get_state_filename (int i) {
	gchar *state_filename_full_path;
	gchar *state_filename;
	gchar *trimlabel;

	/* Test that the CD ROM has been loaded */
	g_return_val_if_fail (strlen (CdromLabel) > 0, NULL);

	/* Remove whitespace from the CdromLabel */
	trimlabel = g_strdup (CdromLabel);
	g_strchomp (trimlabel);

	state_filename = g_strdup_printf ("%s-%.9s.%3.3d", trimlabel, CdromId, i);
	state_filename_full_path = g_build_filename (getenv("HOME"),
												 STATES_DIR,
												 state_filename,
												 NULL);
	printf(" Built state filename %s\n", state_filename_full_path);

	g_free (state_filename);
	g_free (trimlabel);
	
	return state_filename_full_path;
}

void
on_pcsx_configure (GtkWidget *widget, gpointer user_data)
{
	GtkWidget *dlg;
	gint retval;
	
	dlg = pcsx_configure_dialog_new (NULL);
	gtk_dialog_run (GTK_DIALOG (dlg));
	
	/* If all the plugins have been configured, we can enable the relevant menu entries */
	retval = plugins_configured ();
	if (retval)
		pcsx_main_window_set_exec_grp_sensitive (Window, TRUE);

}

void
on_open_bios_menu_activated (GtkWidget *widget, gpointer user_data)
{
	pcsx_launch_psx_bios ();
}

void
on_start_emulation_menu_activated (GtkWidget *widget, gpointer user_data)
{

	/* Get the preference value for whether to show the BIOS loading screen */
	g_object_get (pcsx_gui_prefs, "bios-loadcdbios", &LoadCdBios, NULL);
	
	pcsx_launch_emulation (LoadCdBios);
}


void StartGui() {
	
	Window = pcsx_main_win_new (NULL);
	gtk_widget_show_all (GTK_WIDGET (Window));

	/* If the plugins are not all configured, then
	   disable the emulation-related menu items */
	if (plugins_configured () == FALSE) {
		pcsx_main_window_set_exec_grp_sensitive (Window, FALSE);
	}
	gtk_main ();
	
#ifdef REDUNDANT_CODE	
	GladeXML *xml;
	GtkWidget *widget;
	
	/* If a plugin fails, the Window is not NULL, but is not initialised,
	   so the following causes a segfault
	if (Window != NULL) {
		gtk_window_present (GTK_WINDOW (Window));
		return;
	}*/

	xml = glade_xml_new(PACKAGE_DATA_DIR "pcsx.glade2", "MainWindow", NULL);

	if (!xml)
	{
		g_warning("We could not load the interface!");
		return;
	}

	Window = glade_xml_get_widget(xml, "MainWindow");
	gtk_window_set_title(GTK_WINDOW(Window), "PCSX");
	gtk_window_set_icon_from_file(GTK_WINDOW(Window), PIXMAPDIR "pcsx-icon.png", NULL);
	gtk_window_set_default_icon_from_file(PIXMAPDIR "pcsx-icon.png", NULL);

	ResetMenuSlots(xml);
	
	/* Set up callbacks */
	g_signal_connect_data(GTK_OBJECT(Window), "delete-event",
			GTK_SIGNAL_FUNC(OnDestroy), xml, (GClosureNotify)g_object_unref, G_CONNECT_AFTER);

	/* File menu */
	printf("Processing file menu entries\n");

	widget = glade_xml_get_widget(xml, "loadpsxexemenu");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC (OnFile_RunExe), NULL, NULL, G_CONNECT_AFTER);
	
	widget = glade_xml_get_widget(xml, "loadstatemenu");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC(on_states_load_other), NULL, NULL, G_CONNECT_AFTER);
	
	widget = glade_xml_get_widget(xml, "loadisomenu");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC (on_file_load_cdimage), NULL, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "exit2");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC(OnFile_Exit), NULL, NULL, G_CONNECT_AFTER);
	
	/* Emulation menu */
	printf("Processing emulation menu entries\n");

	widget = glade_xml_get_widget(xml, "loadbiosmenu");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC (on_open_bios_menu_activated), NULL, NULL, G_CONNECT_AFTER);
	
	widget = glade_xml_get_widget(xml, "startemulationmenu");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC (on_start_emulation_menu_activated), NULL, NULL, G_CONNECT_AFTER);
	
	widget = glade_xml_get_widget(xml, "reset1");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC(OnEmu_Reset), NULL, NULL, G_CONNECT_AFTER);

	/* Configuration menu */
	printf("Processing configuration menu entries\n");

	widget = glade_xml_get_widget(xml, "plugins_bios");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC(ConfigurePlugins), NULL, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "plugins_bios");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC(on_pcsx_configure), (gpointer) xml, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "cpu1");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC(OnConf_Cpu), NULL, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "memory_cards");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC(OnConf_Mcds), NULL, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "debug1");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC(OnConf_Dbg), NULL, NULL, G_CONNECT_AFTER);

	/* Help menu */
	printf("Processing help menu entries\n");
	widget = glade_xml_get_widget(xml, "about_pcsx1");
	g_signal_connect_data(GTK_OBJECT(widget), "activate",
			GTK_SIGNAL_FUNC(OnHelp_About), NULL, NULL, G_CONNECT_AFTER);

	/* Set status of menu items that require plugins to have been configured */ 
	if (plugins_configured () == FALSE) {
		/* File menu */
		gtk_widget_set_sensitive (glade_xml_get_widget (xml, "loadpsxexemenu"), TRUE);
		gtk_widget_set_sensitive (glade_xml_get_widget (xml, "loadstatemenu"), TRUE);
		gtk_widget_set_sensitive (glade_xml_get_widget (xml, "loadisomenu"), TRUE);
		
		/* Emulation menu */
		gtk_widget_set_sensitive (glade_xml_get_widget (xml, "loadbiosmenu"), FALSE);
		gtk_widget_set_sensitive (glade_xml_get_widget (xml, "startemulationmenu"), FALSE);
	}
#endif /* REDUNDANT_CODE */
}

#ifdef REDUNDANT_CODE
void OnDestroy() {
	if (!destroy) OnFile_Exit();
}

void ConfigurePlugins() {
	if (!UseGui) {
		/* How do we get here if we're not running the GUI? */
		/* Ryan: we're going to imagine that someday, there will be a way
		 * to configure plugins from the commandline */
		printf("ERROR: Plugins cannot be configured without the GUI.");
		return;
	}

	GladeXML *xml;
	GtkWidget *widget;

	gchar *path;

	ScanPlugins("/usr/lib/games/psemu/");
	ScanPlugins("/usr/local/lib/games/psemu/");
	ScanPlugins(Config.PluginsDir);
	UpdatePluginsBIOS();

	xml = glade_xml_new(PACKAGE_DATA_DIR "pcsx.glade2", "ConfDlg", NULL);

	if (!xml)
	{
		g_warning(_("Error: Glade interface could not be loaded!"));
		return;
	}

	UpdatePluginsBIOS_UpdateGUI(xml);

	ConfDlg = glade_xml_get_widget(xml, "ConfDlg");

	gtk_window_set_title(GTK_WINDOW(ConfDlg), _("Configure PCSX"));

	/* Set the paths in the file choosers to be based on the saved configurations */
	widget = glade_xml_get_widget(xml, "GtkFileChooser_Bios");
	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (widget), 
			Config.BiosDir);
	
	widget = glade_xml_get_widget(xml, "GtkFileChooser_Plugin");
	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (widget), 
			Config.PluginsDir);

	if (strlen(Config.PluginsDir) == 0) {
		if(path = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (widget))) {
			strcpy(Config.PluginsDir, path);
			g_free(path);
		}
	}

	widget = glade_xml_get_widget(xml, "btn_ConfGpu");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(on_configure_plugin), (gpointer) PSE_LT_GPU, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "btn_ConfSpu");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(on_configure_plugin), (gpointer) PSE_LT_SPU, NULL, G_CONNECT_AFTER);

	/* ADB TODO Does pad 1 and 2 need to be different? */
	widget = glade_xml_get_widget(xml, "btn_ConfPad1");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(on_configure_plugin), (gpointer) PSE_LT_PAD, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "btn_ConfPad2");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(on_configure_plugin), (gpointer) PSE_LT_PAD, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "btn_ConfCdr");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(on_configure_plugin), (gpointer) PSE_LT_CDR, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "btn_AboutGpu");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(on_about_plugin), (gpointer) PSE_LT_GPU, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "btn_AboutSpu");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(on_about_plugin), (gpointer) PSE_LT_SPU, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "btn_AboutPad1");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(OnConfConf_Pad1About), xml, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "btn_AboutPad2");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(OnConfConf_Pad2About), xml, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "btn_AboutCdr");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(on_about_plugin), (gpointer) PSE_LT_CDR, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "GtkFileChooser_Bios");
	g_signal_connect_data(GTK_OBJECT(widget), "current_folder_changed",
			GTK_SIGNAL_FUNC(OnBiosPath_Changed), xml, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "GtkFileChooser_Plugin");
	g_signal_connect_data(GTK_OBJECT(widget), "current_folder_changed",
			GTK_SIGNAL_FUNC(OnPluginPath_Changed), xml, NULL, G_CONNECT_AFTER);

	g_signal_connect_data(GTK_OBJECT(ConfDlg), "response",
			GTK_SIGNAL_FUNC(OnConf_Clicked), xml, (GClosureNotify)g_object_unref, G_CONNECT_AFTER);
}
#endif 

static void
destroy_main_window () {
	if (Window) {
        gtk_widget_destroy (GTK_WIDGET (Window));
		Window = NULL;

        gtk_main_quit();
        while (gtk_events_pending()) gtk_main_iteration();
	} else
		printf ("Main window not present - skipping destroy\n");
}

/*
    Launch the emulation from either the command-line or the GUI
	    mode - whether to show the BIOS screen when starting
*/
void
pcsx_launch_emulation (gint mode)
{
	gchar *cdimage;

	destroy_main_window();
	
	LoadPlugins ();
	
	/* If a CD image file has been specified (either on the commandline or
	   in the on_file_load_cdimage () callback, tell the plugin */
	g_object_get (pcsx_gui_prefs, "file-currentfile", &cdimage, NULL);
	if (cdimage) {
		printf ("Setting the CD image file to %s\n", cdimage);
		CDR_setfilename (cdimage);
		g_free (cdimage);
	} else
		printf ("No CD image file specified\n");
	
	if (OpenPlugins() == -1) {
		SysErrorMessage (_("Plugins could not be loaded"),
						 _("The plugins could not be loaded"));
		SysRunGui();
		return;
	}
	
	SysReset();

	/* Only check the CD if we are starting the console with a CD */
	if (CheckCdrom() == -1) {
		ClosePlugins();
		SysErrorMessage (_("CD ROM failed"),
						 _("The CD does not appear to be a valid Playstation CD"));
		SysRunGui();
		return;
	}

	/* Read main executable directly from CDRom and start it */
	if (mode == RUN_CD_BIOS) {
		if (LoadCdrom() == -1) {
			ClosePlugins();
			SysErrorMessage(_("Could not load CD-ROM!"),
							_("The CD ROM could not be loaded"));
			SysRunGui();
			return;
		}
	}

	/* Start emulation */
	SysExecute ();
}

void
pcsx_launch_psx_bios ()
{
	destroy_main_window();

	LoadPlugins ();
	
	/* TODO If a CDR plugin image has not been selected, then OpenPlugins() will prompt us to 
	   configure the plugin, even though we don't need it for the BIOS menu. Need to pass a parameter
	   or something */
	if (OpenPlugins() == -1) {
		/* TODO Error message */
		SysRunGui();
		return;
	}
	
	SysReset();
	
	LoadCdBios = RUN_CD_BIOS;

	SysExecute ();
}

void
pcsx_launch_executable_file (gchar *exefile)
{
	destroy_main_window();
	
	LoadPlugins ();
	
	/* TODO If a CDR plugin image has not been selected, then OpenPlugins() will prompt us to 
	   configure the plugin, even though we don't need it for EXE files. Need to pass a parameter
	   or something */
	if (OpenPlugins() == -1) {
		/* TODO Error message */
		SysRunGui();
	}

	SysReset();
		
	if (Load (exefile) == 0) {
		SysExecute ();
	} else {
		ClosePlugins ();
		SysErrorMessage (_("Not a valid PSX file"),
						 _("The file does not appear to be a valid Playstation executable"));
		SysRunGui ();
	}
}

/* Example return value from filename of format
TEKKEN2-SCES00255.010
is
TEKKEN2
*/
gchar*
get_cdlabel_from_filename (gchar *filename)
{
	gchar **tok;
	gchar *statelabel, *basename;
	
	basename = g_path_get_basename (filename);
	tok = g_strsplit (basename, "-", 0);
	/* Get element on the LHS of the '-' */
	statelabel = g_strdup (tok[0]);

	g_strfreev (tok);
	g_free (basename);
	
	return statelabel;
}

/* Example return value from filename of format
TEKKEN2-SCES00255.010
is
SCES00255
*/
gchar*
get_cdid_from_filename (gchar *filename)
{
	gchar **tok;
	gchar *stateid, *basename;
	
	basename = g_path_get_basename (filename);
	tok = g_strsplit (basename, "-", 0);
	/* Get first 9 characters (i.e. exclude .010) from the RHS of the '-' */
	stateid = g_strdup_printf ("%0.9s", tok[1]);

	g_strfreev (tok);
	g_free (basename);
	
	return stateid;
}

/* Check that the currently loaded CD ROM ID matches that of the CD
   used when saving the state file. The latter is stored in the filename */
gint
validate_state_file_to_cd (gchar *statefile)
{
	gchar *stateid;
	gint retval = 0;

	/* Check the CD ROM is valid */ // AAA FIXME TODO Only do this to get the CDRom label and ID
	if (CheckCdrom() == -1) {
		retval = -1;
	}
	
	stateid = get_cdid_from_filename (statefile);
	printf("Comparing state and CD labels %s and %s\n", stateid, CdromId);
		
	if (g_ascii_strncasecmp (stateid, CdromId, 9) != 0) {
		gchar *error_desc, *statelabel;
			
		statelabel = get_cdlabel_from_filename (statefile);		
			
		error_desc = g_strdup_printf (_("The Playstation CD that is currently in use is %s (%s). "
										"It is not the same CD as that used when saving the state file. "
										"The state file is looking for %s (%s)."),
									  statelabel, stateid,
									  CdromLabel, CdromId);
		SysErrorMessage (_("The CD does not match the state file"),
						 error_desc);

		retval = -1;
		
		g_free (error_desc);
		g_free (statelabel);
	}
	g_free (stateid);
	
	return retval;
}

void
pcsx_launch_state_file (gchar *statefile)
{
	char Text[MAXPATHLEN+20];

	destroy_main_window();
	
	LoadPlugins ();
	
	if (OpenPlugins() == -1) {
		/* TODO Error message */
		SysRunGui();
		return;
	}

	SysReset();

	if (LoadState (statefile) == -1) {
		sprintf(Text, _("Error loading state %s!"), statefile);
		GPU_displayText(Text);
		return;
	}
	
	if (validate_state_file_to_cd (statefile) == 0) {
		sprintf(Text, _("Loaded state %s."), statefile);
		GPU_displayText(Text);

		SysExecute ();
	} else {
		ClosePlugins ();
		SysRunGui ();
	}

}

void
OnFile_RunExe() {
	GtkWidget *file_chooser;

	file_chooser = gtk_file_chooser_dialog_new (_("Select PSX EXE File"),
				NULL, GTK_FILE_CHOOSER_ACTION_OPEN,
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
				NULL);

	/* Add file filter */
	GtkFileFilter *exefilter = gtk_file_filter_new ();
	gtk_file_filter_add_pattern (exefilter, "*.exe");
	gtk_file_filter_add_pattern (exefilter, "*.psx");
	gtk_file_filter_set_name (exefilter, _("Playstation Executable Files"));
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (file_chooser), exefilter);

	GtkFileFilter *allfilter = gtk_file_filter_new ();
	gtk_file_filter_add_pattern (allfilter, "*.*");
	gtk_file_filter_set_name (allfilter, _("All Files"));
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (file_chooser), allfilter);
		
	/* TODO Set this to the config object and retain it - maybe LastUsedDir */
	gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER (file_chooser), getenv("HOME"));

	if (gtk_dialog_run (GTK_DIALOG (file_chooser)) == GTK_RESPONSE_ACCEPT) {
		gchar *file;

		file = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(file_chooser));

		gtk_widget_destroy (file_chooser);
		
		pcsx_launch_executable_file (file);
		g_free (file);

	} else
		gtk_widget_destroy(file_chooser);

}

void
on_file_load_cdimage (GtkWidget *widget, gpointer user_data) {
	GtkWidget *file_chooser;

	file_chooser = gtk_file_chooser_dialog_new (_("Select CD image file"),
				NULL, GTK_FILE_CHOOSER_ACTION_OPEN,
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
				NULL);

	/* TODO Set this to the config object and retain it - maybe LastUsedDir */
	gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER (file_chooser), getenv("HOME"));
	
	if (gtk_dialog_run (GTK_DIALOG (file_chooser)) == GTK_RESPONSE_ACCEPT) {
		gint mode;
		gchar *file;

		file = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(file_chooser));

		gtk_widget_destroy (file_chooser);

		/* Set the current file for use when launching the emulation */
		g_object_set (pcsx_gui_prefs, "file-currentfile", file, NULL);

		/* Get the preference value for whether to show the BIOS loading screen */
		g_object_get (pcsx_gui_prefs, "bios-loadcdbios", &LoadCdBios, NULL);
		
		pcsx_launch_emulation (LoadCdBios);
		
		g_free (file);

	} else
		gtk_widget_destroy(file_chooser);
}

#ifdef REDUNDANT_CODE
static void
reload_plugins () {
	ReleasePlugins();

	if (LoadPlugins() == -1) {
		/* TODO Error handling */
	}
}

void on_file_run_cd (GtkWidget *widget, gpointer user_data) {

	LoadCdBios = (int) user_data;

	destroy_main_window();

	if (OpenPlugins() == -1) {
		/* TODO Error message */
		SysRunGui();
		return;
	}
	
	if (needreset == RESET_REQD) {
		SysReset();
		needreset = RESET_REQD_NO;
	}
printf("Checking = RUN\n");
printf("About to CheckCdrom - LoadCdBios is %d\n", LoadCdBios);
	if (LoadCdBios != RUN) {
		/* Only check the CD if we are starting the console with a CD */
		if (CheckCdrom() == -1) {
			ClosePlugins();
			SysErrorMessage ("CD ROM failed", "The CD does not appear to be a valid Playstation CD");
			SysRunGui();
			return;
		}
	}
printf("Checking = RUN_CD_BIOS\n");
	/* Read main executable directly from CDRom and start it */
	if (LoadCdBios == RUN_CD_BIOS) {
		if (LoadCdrom() == -1) {
			ClosePlugins();
			SysErrorMessage(_("Could not load CD-ROM!"), "The CD ROM could not be loaded");
			SysRunGui();
			return;
		}
	}
	
	if (LoadCdBios == RUN) LoadCdBios = RUN_CD_BIOS;

printf("SysExecute\n");
	SysExecute ();

}
#endif /* REDUNDANT_CODE */

void OnFile_Exit() {
	DIR *dir;
	struct dirent *ent;
	void *Handle;
	gchar *plugin;
	gchar *plugins_dir;
printf ("Exiting PCSX\n");
	g_object_get (pcsx_gui_prefs, "dir-plugins", &plugins_dir, NULL);

	// with this the problem with plugins that are linked with the pthread
	// library is solved

	dir = opendir(plugins_dir);
	if (dir != NULL) {
		while ((ent = readdir(dir)) != NULL) {
			plugin = g_build_filename (plugins_dir, ent->d_name, NULL);

			if (strstr(plugin, ".so") == NULL)
				continue;
			Handle = dlopen(plugin, RTLD_NOW);
			g_free (plugin);
			if (Handle == NULL)
				continue;
		}
	}

	g_free (plugins_dir);

	bind_textdomain_codeset(PACKAGE_NAME, "");
	if (UseGui)
		gtk_main_quit();
	SysClose();
	
	if (UseGui) {
		destroy_main_window ();
		gtk_exit (0);
		
	} else
		exit(0);
}

#ifdef REDUNDANT_CODE
void state_load (gchar *state_filename) {
	int ret;
	char Text[MAXPATHLEN+20];

	/* If the window exists, then we are loading the state from within
	   within the PCSX GUI. We need to initialise the plugins first */
	if (Window) {
		destroy_main_window();

		if (OpenPlugins() == -1) {
			/* TODO Error message */
			SysRunGui();
			return;
		}
	}
	
	if (needreset == RESET_REQD) {
		SysReset();
		needreset = RESET_REQD_NO;
	}

	ret = LoadState(state_filename);
	if (ret == 0) {
		/* Check the CD ROM is valid */
		if (CheckCdrom() == -1) {
			/* TODO Error message */
			ClosePlugins ();
			SysRunGui();
			return;
		}

		/* Check that the currently loaded CD ROM ID matches that of the CD
		   used when saving the state file. The latter is stored in the filename */
		gchar *cmp = g_strrstr (g_path_get_basename (state_filename), "-");
		cmp++;
		if (g_ascii_strncasecmp (cmp, CdromId, 9) != 0) {
			ClosePlugins ();
			gchar *error_desc, *label;
			gint pos;
			label = g_strdup_printf("%.9s", state_filename);
			error_desc = g_strdup_printf (_("The Playstation CD that is currently in use is %s. It is not the same CD as that used when saving the state file. The state file is looking for %s."),
									label,
									CdromLabel);
			SysErrorMessage (_("The CD does not match the state file"),
							 error_desc);
			g_free (error_desc);
			g_free (label);
			SysRunGui();
			return;
		}

		sprintf(Text, _("Loaded state %s."), state_filename);
		GPU_displayText(Text);

		SysExecute ();
	} else {
		sprintf(Text, _("Error loading state %s!"), state_filename);
		GPU_displayText(Text);
	}
}
#endif /* REDUNDANT_CODE */

void state_save (gchar *state_filename) {
	char Text[MAXPATHLEN+20];
	
	GPU_updateLace();

	if (SaveState(state_filename) == 0)
		sprintf(Text, _("Saved state %s."), state_filename);
	else
		sprintf(Text, _("Error saving state %s!"), state_filename);

	GPU_displayText(Text);
}

#ifdef REDUNDANT_CODE
void on_states_load (GtkWidget *widget, gpointer user_data) {
	gchar *state_filename;
	gint state = (int) user_data;
	
	state_filename = get_state_filename (state);
	
	pcsx_launch_state_file (state_filename);
	
	g_free (state_filename);
}

void on_states_save (GtkWidget *widget, gpointer user_data) {
	gchar *state_filename;
	gint state = (int) user_data;
	
	state_filename = get_state_filename (state);

	state_save (state_filename);
	
	g_free (state_filename);
}
#endif /* REDUNDANT_CODE */

void on_states_load_other() {
	GtkWidget *file_chooser;
	gchar *statedir;

	statedir = g_build_path (G_DIR_SEPARATOR_S, getenv ("HOME"), STATES_DIR, NULL);

	file_chooser = gtk_file_chooser_dialog_new (_("Select State File"),
												NULL, GTK_FILE_CHOOSER_ACTION_OPEN,
												GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
												GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
												NULL);
	
	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (file_chooser), statedir);

	if (gtk_dialog_run (GTK_DIALOG (file_chooser)) == GTK_RESPONSE_ACCEPT) {
		gchar *filename;

		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (file_chooser));
		gtk_widget_destroy (file_chooser);

		pcsx_launch_state_file (filename);

		g_free (filename);
	} else
		gtk_widget_destroy (file_chooser);
	
	g_free (statedir);
} 

void on_states_save_other() {
	GtkWidget *file_chooser;
	gchar *statedir;

	statedir = g_build_path (G_DIR_SEPARATOR_S, getenv ("HOME"), STATES_DIR, NULL);

	file_chooser = gtk_file_chooser_dialog_new (_("Select State File"),
												NULL, GTK_FILE_CHOOSER_ACTION_SAVE,
												GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
												GTK_STOCK_SAVE, GTK_RESPONSE_OK,
												NULL);
	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (file_chooser), statedir);

	if (gtk_dialog_run (GTK_DIALOG (file_chooser)) == GTK_RESPONSE_OK) {
		gchar *filename;

		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (file_chooser));
		gtk_widget_destroy(file_chooser);

		state_save (filename);

		g_free (filename);
	}
	else
		gtk_widget_destroy (file_chooser);

	g_free (statedir);
} 

#ifdef REDUNDANT_CODE
void OnEmu_Reset() {
	needreset = RESET_REQD;
}

int all_config_set () {
	int retval;

	if ((strlen(Config.Gpu) != 0) &&
	    (strlen(Config.Spu) != 0) &&
	    (strlen(Config.Cdr) != 0) &&
	    (strlen(Config.Pad1) != 0) &&
	    (strlen(Config.Pad2) != 0))
		retval = TRUE;
	else
		retval = FALSE;

	return retval;
}
#endif /* REDUNDANT_CODE */

#ifdef NETWORKING_ENABLED
GtkWidget *NetDlg;

void OnNet_Clicked (GtkDialog *dialog, gint arg1, gpointer user_data)
{
	if (arg1 == GTK_RESPONSE_OK)
	{
		GetComboText(NetConfS.Combo, NetConfS.plist, Config.Net);

		SaveConfig();
	
		/* ADB Commented out - we should release/load the plugin
		   when we select them, not when we open or close the config screen
		   Either way, net plugin is currently disabled, so this is irrelevant
		ReleasePlugins();
		LoadPlugins();
		needreset = 1;*/
	}

	gtk_widget_destroy(GTK_WIDGET (dialog));
	NetDlg = NULL;
}

void OnConf_Net() {
	GladeXML *xml;
	GtkWidget *widget;

	if (NetDlg != NULL) {
		gtk_window_present (GTK_WINDOW (NetDlg));
		return;
	}

	xml = glade_xml_new(PACKAGE_DATA_DIR "pcsx.glade2", "NetDlg", NULL);

	if (!xml)
	{
		g_warning(_("Error: Glade interface could not be loaded!"));
		return;
	}

	NetDlg = glade_xml_get_widget(xml, "NetDlg");

	FindNetPlugin(xml);

	/* Setup a handler for when Close or Cancel is clicked */
	g_signal_connect_data(GTK_OBJECT(NetDlg), "response",
			GTK_SIGNAL_FUNC(OnNet_Clicked), xml, (GClosureNotify)g_object_unref, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "btn_ConfNet");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(OnNet_Conf), xml, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "btn_ConfAbout");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(OnNet_About), xml, NULL, G_CONNECT_AFTER);
}
#endif /* NETWORKING_ENABLED */

void OnConf_Dbg() {
	printf("STUB for the debugging interface!\n");
}

#ifdef REDUNDANT_CODE
GtkWidget *CpuDlg;
GtkWidget *PsxCombo;
GList *psxglist;
char *psxtypes[] = {
	"NTSC",
	"PAL"
};

/* When the auto-detect CPU type is selected, disable the NTSC/PAL selection */
void OnCpu_PsxAutoClicked (GtkWidget *widget, gpointer user_data)
{
	GtkWidget *combo;
	GladeXML *xml = user_data;
	combo = glade_xml_get_widget(xml, "GtkCombo_PsxType");

	gtk_widget_set_sensitive (combo,
			!(gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))));
}

void OnCpu_Clicked (GtkDialog *dialog, gint arg1, gpointer user_data)
{
	if (arg1 == GTK_RESPONSE_CLOSE)
	{
		GtkWidget *widget;
		GladeXML *xml = user_data;
		int tmp;
		long t;

		widget = glade_xml_get_widget(xml, "GtkCombo_PsxType");
	
		/* If nothing chosen, default to NTSC */
		tmp = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
		if (tmp == -1)	
			tmp = PSX_TYPE_NTSC;
		
		if (!strcmp("NTSC",psxtypes[tmp]))
			Config.PsxType = PSX_TYPE_NTSC;
		else
			Config.PsxType = PSX_TYPE_PAL;

		Config.Xa = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_Xa")));

		Config.Sio = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_Sio")));

		Config.Mdec = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_Mdec")));

		Config.Cdda = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_CDDA")));

		Config.PsxAuto = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_PsxAuto")));

		Config.Dbg = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_Dbg")));

		// Debugger is only working with interpreter not recompiler, so let's set it
		if(Config.Dbg) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_Cpu")), TRUE);
				
		t = Config.Cpu;
		Config.Cpu = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_Cpu")));
		if (t != Config.Cpu) {
			psxCpu->Shutdown();
			#ifdef PSXREC
			if (Config.Cpu)	{
				if(Config.Dbg) psxCpu = &psxIntDbg;
				else psxCpu = &psxInt;
			}
			else psxCpu = &psxRec;
			#else
			if(Config.Dbg) psxCpu = &psxIntDbg;
			else psxCpu = &psxInt;
			#endif
			if (psxCpu->Init() == -1) {
				SysClose();
				exit(1);
			}
			psxCpu->Reset();
		}

		Config.PsxOut = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_PsxOut")));

		Config.SpuIrq = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_SpuIrq")));

		Config.RCntFix = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_RCntFix")));

		Config.VSyncWA = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "GtkCheckButton_VSyncWA")));

		SaveConfig();
	} 

	gtk_widget_destroy(CpuDlg);
	CpuDlg = NULL;
}

void OnConf_Cpu() {
	GladeXML *xml;

	xml = glade_xml_new(PACKAGE_DATA_DIR "pcsx.glade2", "CpuDlg", NULL);

	if (!xml)
	{
		g_warning("We could not load the interface!");
		return;
	}

	CpuDlg = glade_xml_get_widget(xml, "CpuDlg");
	
	PsxCombo = glade_xml_get_widget(xml, "GtkCombo_PsxType");
	gtk_combo_box_set_active (GTK_COMBO_BOX (PsxCombo), Config.PsxType);
	gtk_widget_set_sensitive (GTK_WIDGET (PsxCombo), !Config.PsxAuto);

    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_Xa")), Config.Xa);

    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_Sio")), Config.Sio);

    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_Mdec")), Config.Mdec);

	gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_CDDA")), Config.Cdda);

    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_PsxAuto")), Config.PsxAuto);

	g_signal_connect_data(GTK_OBJECT(glade_xml_get_widget(xml, "GtkCheckButton_PsxAuto")), "toggled",
			GTK_SIGNAL_FUNC(OnCpu_PsxAutoClicked), xml, NULL, G_CONNECT_AFTER);

    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_Cpu")), Config.Cpu);

    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_Dbg")), Config.Dbg);

    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_PsxOut")), Config.PsxOut);

    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_SpuIrq")), Config.SpuIrq);

    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_RCntFix")), Config.RCntFix);

    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (glade_xml_get_widget(xml, "GtkCheckButton_VSyncWA")), Config.VSyncWA);

	/* Setup a handler for when Close or Cancel is clicked */
	g_signal_connect_data(GTK_OBJECT(CpuDlg), "response",
			GTK_SIGNAL_FUNC(OnCpu_Clicked), xml, (GClosureNotify)g_object_unref, G_CONNECT_AFTER);
}


/* TODO Check whether configuration is required when we choose the plugin, and set the state of the
    button appropriately. New gtk tooltip API should allow us to put a tooltip explanation for
    disabled widgets */
/* TODO If combo screen hasn't been opened and the user chooses the menu config option, confs.Combo will be null and cause a segfault */
#define ConfPlugin(src, confs, plugin, name, parent) \
	void *drv; \
	src conf; \
	gchar *filename; \
 \
	GetComboText(confs.Combo, confs.plist, plugin); \
	filename = g_build_filename (getenv("HOME"), PLUGINS_DIR, plugin, NULL); \
	printf("Configuring plugin %s\n", filename); \
	drv = SysLoadLibrary(filename); \
	if (drv == NULL) {printf("Error with file %s\n", filename);return; } \
\
	while (gtk_events_pending()) gtk_main_iteration(); \
	conf = (src) SysLoadSym(drv, name); \
	if (conf) { \
		conf(); \
	} else \
		SysInfoMessage (_("No configuration required"), _("This plugin doesn't need to be configured.")); \
	SysCloseLibrary(drv); \
	g_free (filename);
#endif /* REDUNDANT_CODE */
    
/* TODO This should be done when a plugin is first chosen */
#define TestPlugin(src, confs, plugin, name, parent) \
	void *drv; \
	src conf; \
	int ret = 0; \
	char file[MAXPATHLEN]; \
	gchar *dotdir; \
	gchar *filename; \
 \
	dotdir = g_strconcat (getenv("HOME"), PLUGINS_DIR, NULL); \
 \
	GetComboText(confs.Combo, confs.plist, plugin); \
	filename = g_strconcat (dotdir, plugin); \
	drv = SysLoadLibrary(filename); \
	if (drv == NULL) return; \
 \
	gtk_widget_set_sensitive(parent, FALSE); \
	while (gtk_events_pending()) gtk_main_iteration(); \
	conf = (src) SysLoadSym(drv, name); \
	if (SysLibError() == NULL) ret = conf(); \
	SysCloseLibrary(drv); \
	if (ret == 0) \
		 SysMessage(_("This plugin reports that it should work correctly.")); \
	else SysMessage(_("This plugin reports that it won't work correctly.")); \
	gtk_widget_set_sensitive(parent, TRUE); \
	g_free (dotdir);

#ifdef REDUNDANT_CODE
void OnConfConf_GpuConf(GtkWidget *widget, gpointer user_data) {
	ConfPlugin(GPUconfigure, GpuConfS, Config.Gpu, "GPUconfigure", ConfDlg);
}

void OnConfConf_SpuConf(GtkWidget *widget, gpointer user_data) {
	ConfPlugin(SPUconfigure, SpuConfS, Config.Spu, "SPUconfigure", ConfDlg);
}

void OnConfConf_CdrConf(GtkWidget *widget, gpointer user_data) {
	ConfPlugin(CDRconfigure, CdrConfS, Config.Cdr, "CDRconfigure", ConfDlg);
}

void OnConfConf_Pad1Conf(GtkWidget *widget, gpointer user_data) {
	ConfPlugin(PADconfigure, Pad1ConfS, Config.Pad1, "PADconfigure", ConfDlg);
}

void OnConfConf_Pad2Conf(GtkWidget *widget, gpointer user_data) {
	ConfPlugin(PADconfigure, Pad2ConfS, Config.Pad2, "PADconfigure", ConfDlg);
}

void on_configure_plugin (GtkWidget *widget, gpointer user_data) {
	gint plugin_type = (int) user_data;
	printf("Configuring plugin - %d\n", plugin_type);

	while (gtk_events_pending())
		gtk_main_iteration();
	if (all_config_set() == TRUE) {
		switch (plugin_type) {
			case PSE_LT_GPU:
				OnConfConf_GpuConf (widget, user_data);
				break;
			case PSE_LT_SPU:
				OnConfConf_SpuConf (widget, user_data);
				break;
			case PSE_LT_CDR:
				OnConfConf_CdrConf (widget, user_data);
				break;
			case PSE_LT_PAD:
				OnConfConf_Pad1Conf (widget, user_data);
				break;
#ifdef NETWORKING_ENABLED
			case PSE_LT_NET:
				OnNet_Conf (widget, user_data);
				break;
#endif /* NETWORKING_ENABLED */
		}
	} else
		ConfigurePlugins();
}

void on_about_plugin (GtkWidget *widget, gpointer user_data) {
	gint plugin_type = (int) user_data;
	printf("About plugin - %d\n", plugin_type);
	while (gtk_events_pending())
		gtk_main_iteration();
	if (all_config_set() == TRUE) {
		switch (plugin_type) {
			case PSE_LT_GPU:
				GPU_about();
				break;
			case PSE_LT_SPU:
				SPU_about();
				break;
			case PSE_LT_CDR:
				CDR_about();
				break;
			case PSE_LT_PAD:
				PAD1_about();
				if (strcmp(Config.Pad1, Config.Pad2))
					PAD2_about();
				break;
			case PSE_LT_NET:
				/* ADB TODO */
				break;
		}
	} else
		ConfigurePlugins();
}

void OnConfConf_Pad1About(GtkWidget *widget, gpointer user_data) {
	ConfPlugin(PADabout, Pad1ConfS, Config.Pad1, "PADabout", ConfDlg);
}

void OnConfConf_Pad2About(GtkWidget *widget, gpointer user_data) {
	ConfPlugin(PADabout, Pad2ConfS, Config.Pad2, "PADabout", ConfDlg);
}
#endif /* REDUNDANT_CODE */

#ifdef NETWORKING_ENABLED
void OnNet_Conf(GtkWidget *widget, gpointer user_data) {
	ConfPlugin(NETconfigure, NetConfS, Config.Net, "NETconfigure", NetDlg);
}

void OnNet_About(GtkWidget *widget, gpointer user_data) {
	ConfPlugin(NETabout, NetConfS, Config.Net, "NETabout", NetDlg);
}
#endif /* NETWORK_ENABLED */

#ifdef REDUNDANT_CODE
void PickDefault (char* confval, char* currchosenval) {;
	/* If no value is set in the config file, then pick the first value in the available plugin list */
	/* TODO Set this when the plugins are first added to the combo list instead */
	if ((currchosenval) && (strlen(confval) == 0)) {
		strcpy(confval, currchosenval);
		printf("Picking default plugin: %s\n", currchosenval);
	}
}

void OnPluginPath_Changed(GtkWidget *wdg, gpointer data) {
	gchar *path;

	path = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (wdg));
	strcpy(Config.PluginsDir, path);

	plugin_file_handler_symlink_plugins (Config.PluginsDir);
	UpdatePluginsBIOS();

	// Pick some defaults, if they're available and no Config.x is set
	/* ADB TODO Same as set_default_plugin in LnxMain.c!! */
	PickDefault (Config.Gpu, GpuConfS.plist[0]);
	PickDefault (Config.Spu, SpuConfS.plist[0]);
	PickDefault (Config.Cdr, CdrConfS.plist[0]);
	PickDefault (Config.Pad1, Pad1ConfS.plist[0]);
	PickDefault (Config.Pad2, Pad2ConfS.plist[0]);
	PickDefault (Config.Bios, BiosConfS.plist[0]);
	
	UpdatePluginsBIOS_UpdateGUI(data);
	g_free (path);
}

void OnBiosPath_Changed(GtkWidget *wdg, gpointer data) {

	gchar *foldername;

	foldername = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (wdg));
	strcpy(Config.BiosDir, foldername);

	printf("BIOS directory is now %s\n", foldername);
	
	UpdatePluginsBIOS();
	UpdatePluginsBIOS_UpdateGUI(data);
	
	g_free (foldername);
}

void OnConf_Clicked (GtkDialog *dialog, gint arg1, gpointer user_data)
{

	if (arg1 == GTK_RESPONSE_OK)
	{
		GetComboText(GpuConfS.Combo, GpuConfS.plist, Config.Gpu);
		GetComboText(SpuConfS.Combo, SpuConfS.plist, Config.Spu);
		GetComboText(CdrConfS.Combo, CdrConfS.plist, Config.Cdr);
		GetComboText(Pad1ConfS.Combo, Pad1ConfS.plist, Config.Pad1);
		GetComboText(Pad2ConfS.Combo, Pad2ConfS.plist, Config.Pad2);
		GetComboText(BiosConfS.Combo, BiosConfS.plist, Config.Bios);
		/* TODO Validation */

		SaveConfig();

		/* ADB Commented out - we should release/load the plugin
		   when we select them, not when we open or close the config screen
		*/
		ReleasePlugins();

		// Ryan: inelegant hack to stop the 'Could not load' error.
		// if all plugins aren't available, don't try to load them.
		if ((strlen(Config.Gpu) == 0) && (strlen(Config.Spu) == 0) && (strlen(Config.Cdr) == 0) && (strlen(Config.Pad1) == 0) && (strlen(Config.Pad2) == 0) && (strlen(Config.Bios) == 0)) {
			LoadPlugins();		/* TODO Check for return code -1 and handle */
		}

		needreset = RESET_REQD;
	} 
	gtk_widget_destroy (ConfDlg);
	ConfDlg = NULL;

	if (all_config_set() == TRUE)
		reload_plugins();
}
#endif /* REDUNDANT_CODE */

const gchar *authors [] = { 
	"Ryan Schultz <schultz.ryan@gmail.com>",
	"Andrew Burton <adb@iinet.net.au>",
	"Stephen Chao <stephen@digitalnexus.org>",
	"Stefan Sikora <hoshy[at]schrauberstube.de>",
	"PCSX Team",
	NULL
};

const gchar *artists [] = {
	"Ryan Schultz <schultz.ryan@gmail.com>",
	NULL
};

const gchar *documenters [] = {
	"Ryan Schultz <schultz.ryan@gmail.com>",
	NULL
};

const gchar *copyright =
	"(C) 1999-2003  PCSX Team\n"
	"(C) 2005-2006 Ryan Schultz\n"
	"(C) 2005-2009 Andrew Burton\n"
	"(C) 2008-2009 Stefan Sikora\n";

void OnHelp_About(GtkWidget *widget, gpointer user_data) {
	GtkWidget *about_dialog;

	about_dialog = gtk_about_dialog_new ();
	gtk_about_dialog_set_program_name (GTK_ABOUT_DIALOG (about_dialog), "PCSX-df");
	gtk_about_dialog_set_version (GTK_ABOUT_DIALOG (about_dialog), VERSION);
	gtk_about_dialog_set_website (GTK_ABOUT_DIALOG (about_dialog), "http://sourceforge.net/projects/pcsx-df/");
	gtk_about_dialog_set_website_label (GTK_ABOUT_DIALOG (about_dialog), _("PCSX-df Development"));
	gtk_about_dialog_set_authors (GTK_ABOUT_DIALOG (about_dialog), authors);
	gtk_about_dialog_set_copyright (GTK_ABOUT_DIALOG (about_dialog), copyright);
	gtk_about_dialog_set_documenters (GTK_ABOUT_DIALOG (about_dialog), documenters);
	gtk_about_dialog_set_artists (GTK_ABOUT_DIALOG (about_dialog), artists);
	gtk_about_dialog_set_translator_credits (GTK_ABOUT_DIALOG (about_dialog), _("translator-credits"));
	gtk_about_dialog_set_comments (GTK_ABOUT_DIALOG (about_dialog), _("A Playstation emulator for GNU/Linux."));
	gtk_about_dialog_set_license (GTK_ABOUT_DIALOG (about_dialog),
								  _("PCSX is licensed under the GNU GPL; see the included COPYING file for more detail."));
	gtk_about_dialog_set_wrap_license (GTK_ABOUT_DIALOG (about_dialog), TRUE);
								  
	gtk_dialog_run (GTK_DIALOG (about_dialog));
	gtk_widget_destroy (about_dialog);
}

#ifdef REDUNDANT_CODE
#define ComboAddPlugin(type) { \
	type##ConfS.plugins+=2; \
	strcpy(type##ConfS.plist[type##ConfS.plugins-1], name); \
	strcpy(type##ConfS.plist[type##ConfS.plugins-2], ent->d_name); \
	type##ConfS.glist = g_list_append(type##ConfS.glist, type##ConfS.plist[type##ConfS.plugins-1]); \
}

void populate_combo_box (GtkWidget *widget, GList *list) {
	GtkListStore *store;
	GtkCellRenderer *renderer;
	store = gtk_list_store_new (1, G_TYPE_STRING);

	/* Clear existing data from combo box */
	gtk_cell_layout_clear (GTK_CELL_LAYOUT (widget));

	renderer = gtk_cell_renderer_text_new();
        gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (widget), renderer, FALSE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (widget), renderer, "text", 0);
		
	while (list != NULL) {
		GtkTreeIter iter;
		gtk_list_store_append (store, &iter);
		gtk_list_store_set (store, &iter, 0, (char *)list->data, -1);
		list = list->next;
	}
	gtk_combo_box_set_model (GTK_COMBO_BOX (widget), GTK_TREE_MODEL (store));
}

#define ConfCreatePConf(name, type) \
	/* Populate the relevant combo widget with the list of plugins. \
	   If no plugins available, disable the combo and its controls.
	   Note that the Bios plugin has no About/Conf control. */ \
	type##ConfS.Combo = glade_xml_get_widget(xml, "GtkCombo_" name); \
	if (type##ConfS.glist != NULL) { \
		populate_combo_box (type##ConfS.Combo, type##ConfS.glist); \
		FindComboText(type##ConfS.Combo, type##ConfS.plist, Config.type); \
		gtk_widget_set_sensitive (type##ConfS.Combo, TRUE); \
		if (g_ascii_strcasecmp (name, "Bios") != 0) { \
			controlwidget = glade_xml_get_widget(xml, "btn_Conf" name); \
			gtk_widget_set_sensitive (controlwidget, TRUE); \
			controlwidget = glade_xml_get_widget(xml, "btn_About" name); \
			gtk_widget_set_sensitive (controlwidget, TRUE); \
		} \
	} else { \
		if (g_ascii_strcasecmp (name, "Bios") != 0) { \
			gtk_widget_set_sensitive (type##ConfS.Combo, FALSE); \
			controlwidget = glade_xml_get_widget(xml, "btn_Conf" name); \
			gtk_widget_set_sensitive (controlwidget, FALSE); \
			controlwidget = glade_xml_get_widget(xml, "btn_About" name); \
			gtk_widget_set_sensitive (controlwidget, FALSE); \
		} \
	}
#endif /* REDUNDANT_CODE */

void scan_bios_dir () {
	gchar *bios_dir;
	DIR *dir;
	struct dirent *ent;
	gchar *filename;

	/* Destroy any existing GList */
	bios_list = NULL;
	
	/* Add Internal HLE bios */
	pcsx_bios *bios;
	bios = g_new0 (pcsx_bios, 1);
	bios->display_name = g_strdup (_("Internal HLE Bios"));
	bios->internal_name = g_strdup (_("HLE"));
									
	bios_list = g_list_append (bios_list, bios);

	g_object_get (pcsx_gui_prefs, "dir-bios", &bios_dir, NULL);
	
	/* printf("Scanning bios dir %s\n", bios_dir); */

	dir = opendir (bios_dir);
	g_return_if_fail (dir != NULL);

	while ((ent = readdir(dir)) != NULL) {
		filename = g_build_filename (bios_dir, ent->d_name, NULL);

		if (is_valid_bios_file (filename)) {

			pcsx_bios *bios;
			bios = g_new0 (pcsx_bios, 1);
			bios->display_name = g_strdup (g_path_get_basename (filename));
			bios->internal_name = g_strdup (g_path_get_basename (filename));
			
			bios_list = g_list_append (bios_list, bios);
		}
	}
	closedir(dir);
	/* printf("Finished scanning bios dir %s\n", bios_dir); */
	
	g_free (filename);
	g_free (bios_dir);

}

#ifdef REDUNDANT_CODE
void check_symlinks_in_path (char* dotdir)
{
	DIR *dir;
	struct dirent *ent;
	struct stat stbuf;
	gchar *linkname;

	dir = opendir(dotdir);
	if (dir == NULL) {
		SysMessage(_("Could not open directory: '%s'\n"), dotdir);
		return;
	}

	/* Check for any bad links in the directory. If the remote
	   file no longer exists, remove the link */
	while ((ent = readdir(dir)) != NULL) {
		linkname = g_strconcat (dotdir, ent->d_name, NULL);

		if (stat(linkname, &stbuf) == -1) {
			/* File link is bad, remove it */
			printf("unlink: %s\n", linkname);
			unlink(linkname);
		}
		g_free (linkname);
	}
	closedir(dir);
}
#endif /*REDUNDANT_CODE */



int is_valid_bios_file (gchar *filename) {
	int valid;
	struct stat buf;

	/* printf("Checking if %s is a valid BIOS file", filename); */
	if ((stat(filename, &buf) == -1) || (buf.st_size != (1024*512)))
		valid = FALSE;
	else {
		valid = TRUE;
		/* printf(" ...yes\n"); */
	}

	return valid;
}

#ifdef REDUNDANT_CODE
/* Add the name of the BIOS file to the drop-down list. This will
   be the filename, not the full path to the file */
void add_bios_to_list (gchar *bios_name, gchar *internal_name) {
	BiosConfS.plugins+=2;
	strcpy(BiosConfS.plist[BiosConfS.plugins-1], bios_name);
	strcpy(BiosConfS.plist[BiosConfS.plugins-2], internal_name);
	/* Error handling - BIOS
	strcpy(BiosConfS.plist[BiosConfS.plugins-1], internal_name);
	strcpy(BiosConfS.plist[BiosConfS.plugins-2], bios_name);*/
	BiosConfS.glist = g_list_append(BiosConfS.glist, BiosConfS.plist[BiosConfS.plugins-1]);
}


void scan_bios_dir (gchar *dirname) {
	DIR *dir;
	struct dirent *ent;
	gchar *filename;

	printf("Scanning bios dir %s\n", dirname);

	dir = opendir(dirname);
	if (dir == NULL) {
		SysMessage(_("Could not open BIOS directory: '%s'\n"), dirname);
		return;
	}

	while ((ent = readdir(dir)) != NULL) {
		filename = g_build_filename (dirname, ent->d_name, NULL);
		if (is_valid_bios_file (filename))
			add_bios_to_list (g_path_get_basename (filename), g_path_get_basename (filename));
	}
	closedir(dir);
	g_free (filename);

	printf("Finished scanning bios dir %s\n", dirname);
}

void UpdatePluginsBIOS() {
	DIR *dir;
	struct dirent *ent;
	struct stat stbuf;
	void *Handle;
	char name[256];
	gchar *linkname;
	gchar *dotpluginsdir;
	gchar *dotpluginscfgdir;
	gchar *dotdir;

	GpuConfS.plugins  = 0; SpuConfS.plugins  = 0; CdrConfS.plugins  = 0;
	Pad1ConfS.plugins = 0; Pad2ConfS.plugins = 0; BiosConfS.plugins = 0;
	GpuConfS.glist  = NULL; SpuConfS.glist  = NULL; CdrConfS.glist  = NULL;
	Pad1ConfS.glist = NULL; Pad2ConfS.glist = NULL; BiosConfS.glist = NULL;

	/* Check for bad links in ~/.pcsx/plugins/ */
	dotpluginsdir = g_build_filename (getenv("HOME"), PLUGINS_DIR, NULL);
	check_symlinks_in_path (dotpluginsdir);
	
	/* Check for bad links in ~/.pcsx/plugins/cfg */
	dotpluginscfgdir = g_build_filename (getenv("HOME"), PLUGINS_CFG_DIR, NULL);
	check_symlinks_in_path (dotpluginscfgdir);

	/* Load and get plugin info */
	dir = opendir(dotpluginsdir);
	if (dir == NULL) {
		printf(_("Could not open plugins directory: '%s'\n"), dotpluginsdir);
		return;
	}
	while ((ent = readdir(dir)) != NULL) {
		long type,v;
		linkname = g_build_filename (dotpluginsdir, ent->d_name, NULL);
		
		// only libraries past this point, not config tools
		if (strstr(linkname, ".so") == NULL) continue;

		Handle = dlopen(linkname, RTLD_NOW);
		if (Handle == NULL) { printf("%s\n", dlerror()); continue; }

		PSE_getLibType = (PSEgetLibType) dlsym(Handle, "PSEgetLibType");
		if (dlerror() != NULL) {
			if (strstr(linkname, "gpu") != NULL) type = PSE_LT_GPU;
			else if (strstr(linkname, "cdr") != NULL) type = PSE_LT_CDR;
			else if (strstr(linkname, "spu") != NULL) type = PSE_LT_SPU;
			else if (strstr(linkname, "pad") != NULL) type = PSE_LT_PAD;
			else continue;
		}
		else type = PSE_getLibType();

		PSE_getLibName = (PSEgetLibName) dlsym(Handle, "PSEgetLibName");
		if (dlerror() == NULL) {
			sprintf(name, "%s", PSE_getLibName());
			PSE_getLibVersion = (PSEgetLibVersion) dlsym(Handle, "PSEgetLibVersion");
			if (dlerror() == NULL) {
				char ver[32];

				v = PSE_getLibVersion();
				sprintf(ver, " %ld.%ld.%ld",v>>16,(v>>8)&0xff,v&0xff);
				strcat(name, ver);
			}
		}
		else strcpy(name, ent->d_name);

		if (type & PSE_LT_CDR)
			ComboAddPlugin(Cdr);
		if (type & PSE_LT_GPU)
			ComboAddPlugin(Gpu);
		if (type & PSE_LT_SPU)
			ComboAddPlugin(Spu);
		if (type & PSE_LT_PAD) {
			PADquery query = (PADquery)dlsym(Handle, "PADquery");
			if (query() & 0x1) {
				ComboAddPlugin(Pad1);
			}
			if (query() & 0x2) {
				ComboAddPlugin(Pad2);
			}
		}
	}
	closedir(dir);

	/* The BIOS list always contains the PCSX internal BIOS */
	add_bios_to_list (_("Internal HLE Bios"), "HLE");

	scan_bios_dir (Config.BiosDir);

	/* also scan the .pcsx/bios/ directory */
	//dotdir = g_build_filename (getenv("HOME"), BIOS_DIR, NULL);
	//scan_bios_dir (dotdir);

	g_free (dotpluginsdir);
	g_free (dotpluginscfgdir);
	//g_free (dotdir);
	g_free (linkname);
}

void UpdatePluginsBIOS_UpdateGUI(GladeXML *xml) {
	/* Populate the plugin combo boxes */
	ConfCreatePConf("Gpu", Gpu);
	ConfCreatePConf("Spu", Spu);
	ConfCreatePConf("Pad1", Pad1);
	ConfCreatePConf("Pad2", Pad2);
	ConfCreatePConf("Cdr", Cdr);
	ConfCreatePConf("Bios", Bios);
}
#endif /* REDUNDANT_CODE */

#ifdef NETWORKING_ENABLED
void FindNetPlugin(GladeXML *xml) {
	DIR *dir;
	struct dirent *ent;
	void *Handle;
	char plugin[MAXPATHLEN],name[MAXPATHLEN];
	gchar *dotdir;
	
	dotdir = g_strconcat (getenv("HOME"), BIOS_DIR, NULL);

	NetConfS.plugins  = 0;
	NetConfS.glist = NULL; 

	NetConfS.plugins+=2;
	strcpy(NetConfS.plist[NetConfS.plugins-1], "Disabled");
	strcpy(NetConfS.plist[NetConfS.plugins-2], "Disabled");
	NetConfS.glist = g_list_append(NetConfS.glist, NetConfS.plist[NetConfS.plugins-1]);

	dir = opendir(dotdir);
	if (dir == NULL)
		SysMessage(_("Could not open directory: '%s'\n"), dotdir);
	else {

		/* ADB TODO Replace the following with a function */
		while ((ent = readdir(dir)) != NULL) {
			long type,v;

			sprintf(plugin, "%s%s", dotdir, ent->d_name);

			if (strstr(plugin, ".so") == NULL) continue;
			Handle = dlopen(plugin, RTLD_NOW);
			if (Handle == NULL) continue;

			PSE_getLibType = (PSEgetLibType) dlsym(Handle, "PSEgetLibType");
			if (dlerror() != NULL) {
				if (strstr(plugin, "net") != NULL) type = PSE_LT_NET;
				else continue;
			}
			else type = PSE_getLibType();

			PSE_getLibName = (PSEgetLibName) dlsym(Handle, "PSEgetLibName");
			if (dlerror() == NULL) {
				sprintf(name, "%s", PSE_getLibName());
				PSE_getLibVersion = (PSEgetLibVersion) dlsym(Handle, "PSEgetLibVersion");
				if (dlerror() == NULL) {
					char ver[32];

					v = PSE_getLibVersion();
					sprintf(ver, " %ld.%ld.%ld",v>>16,(v>>8)&0xff,v&0xff);
					strcat(name, ver);
				}
			}
			else strcpy(name, ent->d_name);

			if (type & PSE_LT_NET) {
				ComboAddPlugin(Net);
			}
		}
		closedir(dir);

		ConfCreatePConf("Net", Net);
	}
	g_free (dotdir);
}
#endif /* NETWORKING_ENABLED */

void SysMessage(char *fmt, ...) {
	GtkWidget *dlg;
	va_list list;
	char msg[512];
	
	/* Format the message that was passed in */
	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);

	if (msg[strlen(msg)-1] == '\n') msg[strlen(msg)-1] = 0;
	
	if (!UseGui)
		printf ("MESSAGE: %s\n",msg);
	else {
		dlg = gtk_message_dialog_new (NULL,
									  GTK_DIALOG_MODAL,
									  GTK_MESSAGE_WARNING,
									  GTK_BUTTONS_OK,
									  _("An error has occurred during emulation"),
									  NULL);
										 
		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dlg),
												  msg);
	
		gtk_dialog_run (GTK_DIALOG (dlg));
		gtk_widget_destroy (dlg);
	}
}

void SysErrorMessage (gchar *primary, gchar *secondary) {
	GtkWidget *message_dialog;	
	if (!UseGui)
		printf ("ERROR: %s - %s\n", primary, secondary);
	else {
		message_dialog =  gtk_message_dialog_new (NULL,
				GTK_DIALOG_MODAL,
				GTK_MESSAGE_ERROR,
				GTK_BUTTONS_CLOSE,
				primary,
				NULL);
		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (message_dialog),
				secondary);
	
		gtk_dialog_run (GTK_DIALOG (message_dialog));
		gtk_widget_destroy (message_dialog);
	}
}

void SysInfoMessage  (gchar *primary, gchar *secondary) {
	GtkWidget *message_dialog;	
	if (!UseGui)
		printf ("INFO: %s - %s\n", primary, secondary);
	else {
		message_dialog =  gtk_message_dialog_new (NULL,
				GTK_DIALOG_MODAL,
				GTK_MESSAGE_INFO,
				GTK_BUTTONS_CLOSE,
				primary,
				NULL);
		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (message_dialog),
				secondary);
	
		gtk_dialog_run (GTK_DIALOG (message_dialog));
		gtk_widget_destroy (message_dialog);
	}
}
