/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <dlfcn.h>
#include <gtk/gtk.h>

#include "Linux.h"
#include "Config.h" /* Gui Prefs */
#include "pcsx-memcard-handler.h"
/* FIXME */
#include "../libpcsxcore/sio.h"

#ifdef ENABLE_NLS
#include <locale.h>
#endif

enum {
	DONT_USE_GUI,
	USE_GUI
};

int UseGui = USE_GUI;

#ifdef REDUNDANT_CODE
/* Set the default plugin name */
void set_default_plugin (char *plugin_name, char *conf_plugin_name) {
	if (strlen(plugin_name) != 0) {
		strcpy(conf_plugin_name, plugin_name);
		printf("Picking default plugin: %s\n", plugin_name);
	} else
		printf("No default plugin could be found for %s\n", conf_plugin_name);
	
}
#endif /* REDUNDANT_CODE */

int main(int argc, char *argv[]) {
	int runcd = 0;
	gint loadst = 0;
	gchar *plugin_dir;

#ifdef ENABLE_NLS
	setlocale (LC_ALL, "");
	bindtextdomain (GETTEXT_PACKAGE, LOCALE_DIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);
#endif

	static gboolean nousegui = FALSE;
	static gchar *filename_exe = NULL;
	static gchar *filename_cd = NULL;
	static gint runcda;
	static gint runcdbios;
	
	GOptionContext *ctx;
	GOptionEntry options[] = {
		{ "runcd", 0, 0, G_OPTION_ARG_NONE, &runcda, "Run CD", NULL  },
		{ "runcdbios", 0, 0, G_OPTION_ARG_NONE, &runcdbios, "Run CD BIOS", NULL  },
		{ "nogui", 0, 0, G_OPTION_ARG_NONE, &nousegui, "Run PCSX emulation from the command line with no GUI", NULL },
		{ "psxout", 0, 0, G_OPTION_ARG_NONE, &Config.PsxOut, "Enable PSX output", NULL },
		{ "load", 0, 0, G_OPTION_ARG_INT, &loadst, "Loads savestate STATENUM (1-5)", "STATENUM" },
		{ "exefile", 0, 0, G_OPTION_ARG_FILENAME, &filename_exe, "Loads PSX EXE", "FILE" },
		{ "cdfile", 0, 0, G_OPTION_ARG_FILENAME, &filename_cd, "Loads PSX CD file", "FILE" },
		{ NULL }
	};
	
	ctx = g_option_context_new("- Playstation emulator");
	g_option_context_add_main_entries(ctx, options, "pcsx");
	g_option_context_parse(ctx, &argc, &argv, NULL);
	g_option_context_free(ctx);
	
	/* Set options based on the command line parameters */
	if (nousegui) UseGui = DONT_USE_GUI;
	
	if (runcda) runcd = RUN_CD;
	
	if (runcdbios) runcd = RUN_CD_BIOS;
	
	if (filename_cd) runcd = RUN_CD;
	
	memset(&Config, 0, sizeof(PcsxConfig));
#ifdef NETWORKING_ENABLED
	strcpy(Config.Net, _("Disabled"));
#endif NETWORKING_ENABLED

	/* Need to initialise GTK even if running without a GUI to load the GObjects */
	gtk_init (NULL, NULL);
	
	/* Look to make sure the .pcsx subdirectories are available */
	check_sub_directories ();
	
	pcsx_gui_prefs = pcsx_gui_prefs_new ();

	g_object_get (pcsx_gui_prefs, "dir-plugins", &plugin_dir, NULL);
	//plugin_file_handler_symlink_plugins ("/usr/lib/games/psemu/");
	plugin_file_handler_symlink_plugins (plugin_dir);
	g_free (plugin_dir);

#ifdef REDUNDANT_CODE
        // Uh oh, no config file found, use some defaults
		Config.PsxAuto = 1;	/* ADB TODO */

		gchar *str_bios_dir = g_strconcat (getenv("HOME"), "/.pcsx/bios/", NULL);
		strcpy(Config.BiosDir,  str_bios_dir);
		g_free (str_bios_dir);

		// scan some default locations to find plugins
		// ADB TODO These are Linux-specific - what about Windows?
		gchar *currentdir;
		ScanPlugins("/usr/lib/games/psemu/");
		ScanPlugins("/usr/lib/games/psemu/lib/");
		ScanPlugins("/usr/lib/games/psemu/config/");
		ScanPlugins("/usr/local/lib/games/psemu/lib/");
		ScanPlugins("/usr/local/lib/games/psemu/config/");
		ScanPlugins("/usr/local/lib/games/psemu/");
		
		currentdir = g_strconcat (getenv("HOME"), "/.psemu-plugins/", NULL);
		ScanPlugins(currentdir);

		currentdir = g_strconcat (getenv("HOME"), "/.psemu/", NULL);
		ScanPlugins(currentdir);
		g_free (currentdir);
		
		// Update available plugins, but not GUI
		UpdatePluginsBIOS();

		// Pick some defaults, if they're available
		set_default_plugin (GpuConfS.plist[0], Config.Gpu);
		set_default_plugin (SpuConfS.plist[0], Config.Spu);
		set_default_plugin (CdrConfS.plist[0], Config.Cdr);
		set_default_plugin (Pad1ConfS.plist[0], Config.Pad1);
		set_default_plugin (Pad2ConfS.plist[0], Config.Pad2);
		set_default_plugin (BiosConfS.plist[0], Config.Bios);
#endif /* REDUNDANT_CODE */

	/* Create & load default memcards if they don't exist */
	/* FIXME TODO Move this to when the memcard dialog opens, or when emulation starts. There
	   is no point creating/loading/freeing them here */
	gchar *memcard1, *memcard2;
	g_object_get (pcsx_gui_prefs,
				  "memcard-memcard1", &memcard1,
				  "memcard-memcard2", &memcard2,
				  NULL);
	create_memory_card ("card1.mcd", memcard1);
	create_memory_card ("card2.mcd", memcard2);
printf("Created memory cards\n");
	LoadMcds(memcard1, memcard2);
printf("Loaded memory cards\n");
	g_free (memcard1);
	g_free (memcard2);

#ifdef REDUNDANT_CODE
		SaveConfig();
#endif /* REDUNDANT_CODE */
	
	/* Build a list of plugins and BIOSes */
	scan_plugins_dir ();
	scan_bios_dir ();

	if (SysInit() == -1)
		return -1;

	if (UseGui == USE_GUI)
		StartGui();
	else {
		
		/* Set the file so it can be set after the CDR plugin has been loaded
		   using CDR_setfilename() */
		if (filename_cd)
			g_object_set (pcsx_gui_prefs, "file-currentfile", filename_cd, NULL);

		
		/* The following only occurs if the gui isn't started */
		if (filename_exe) {
			/* Load EXE file */
			printf("Loading PSX EXE file %s\n", filename_exe);
			pcsx_launch_executable_file (filename_exe);
			return;
		}
		
		/* If a state has been specified, then load that */
		if (loadst > 0) {
			StatesC = loadst-1;
			printf("Loading state %d\n", StatesC);
			gchar *state_filename = get_state_filename (StatesC);
			g_return_if_fail (state_filename != NULL);
			
			pcsx_launch_state_file (state_filename);

			g_free (state_filename);
			return;
		}
		
		/* Default - if not loading EXE or a state then we are either loading
		   an ISO file or launching emulation to the BIOS screen */
		LoadCdBios = runcd;
		pcsx_launch_emulation (LoadCdBios);
	}
	return 0;
}

int SysInit() {

	gchar *memcard1, *memcard2;
	
#ifdef GTE_DUMP
	gteLog = fopen("gteLog.txt","wb");
	setvbuf(gteLog, NULL, _IONBF, 0);
#endif

#ifdef EMU_LOG
#ifndef LOG_STDOUT
	emuLog = fopen("emuLog.txt","wb");
#else
	emuLog = stdout;
#endif
	setvbuf(emuLog, NULL, _IONBF, 0);
#endif

	if (psxInit() == -1) {
		/* TODO Error handling */
		printf("PSX emulator couldn't be initialised.\n");
		return -1;
	}

	g_object_get (pcsx_gui_prefs,
				  "memcard-memcard1", &memcard1,
				  "memcard-memcard2", &memcard2,
				  NULL);
	LoadMcds(memcard1, memcard2);
	g_free (memcard1);
	g_free (memcard2);

	return 0;
}

void SysReset() {
	psxReset();
}

void SysExecute() {
	if(Config.Dbg)
		hdb_start();

	psxCpu->Execute();
}

void SysClose() {
	psxShutdown();
	ReleasePlugins();

	if (emuLog != NULL) fclose(emuLog);
}

void SysPrintf(char *fmt, ...) {
	va_list list;
	char msg[512];

	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);

	if (Config.PsxOut) printf ("%s", msg);
#ifdef EMU_LOG
#ifndef LOG_STDOUT
	fprintf(emuLog, "%s", msg);
#endif
#endif
}

void *SysLoadLibrary(char *lib) {
	return dlopen(lib, RTLD_NOW);
}

void *SysLoadSym(void *lib, char *sym) {
	return dlsym(lib, sym);
}

const char *SysLibError() {
	return dlerror();
}

void SysCloseLibrary(void *lib) {
	dlclose(lib);
}

void SysUpdate() {
	PADhandleKey(PAD1_keypressed());
	PADhandleKey(PAD2_keypressed());
}

void SysRunGui() {
	StartGui();
}
