/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2008  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gtk/gtk.h>
#include <glade/glade.h>

#include "Linux.h"
#include "Config.h"
#include "../libpcsxcore/sio.h"

#include "memcard-dlg.h"

#define MAX_MEMCARD_BLOCKS 15

GTimer *Gtimer;
int timer;
McdBlock Blocks[2][MAX_MEMCARD_BLOCKS];	/* Assuming 2 cards, 15 blocks? */
int IconC[2][MAX_MEMCARD_BLOCKS];
enum {
    CL_ICON,
    CL_TITLE,
    CL_STAT,
    CL_ID,
    CL_NAME,
    NUM_CL
};

GtkWidget *GtkCList_McdList1, *GtkCList_McdList2;
GtkTreeSelection *sel1, *sel2;
gint mcd1_row, mcd2_row;

static void
copy_memcard_data (char *from, char *to, gint *i, gchar *str);
static int
get_free_memcard_slot (int target_card);

static void
add_columns (GtkTreeView *treeview) {
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;

	/* column for icon */
	renderer = gtk_cell_renderer_pixbuf_new ();
	column = gtk_tree_view_column_new_with_attributes (_("Icon"),
			renderer, "pixbuf", CL_ICON, NULL);
	gtk_tree_view_append_column (treeview, column);

	/* column for title */
	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes (_("Title"),
			renderer, "text", CL_TITLE, NULL);
	gtk_tree_view_append_column (treeview, column);

	/* column for status */
	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes (_("Status"),
			renderer, "text", CL_STAT, NULL);
	gtk_tree_view_append_column (treeview, column);

	/* column for id */
	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes (_("ID"),
			renderer, "text", CL_ID, NULL);
	gtk_tree_view_append_column (treeview, column);

	/* column for Name */
	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes (_("Name"),
			renderer, "text", CL_NAME, NULL);
	gtk_tree_view_append_column (treeview, column);							    
}

static GdkPixbuf
*SetIcon(GtkWidget *dialog, short *icon, int i) {
	GdkPixmap *pixmap;
	GdkImage  *image;
	GdkVisual *visual;
	GdkGC     *gc;
	int x, y, c;

	visual = gdk_window_get_visual(dialog->window);

	if (visual->depth == 8) return NULL;

	image = gdk_image_new(GDK_IMAGE_NORMAL, visual, 16, 16);

	for (y=0; y<16; y++) {
		for (x=0; x<16; x++) {
			c = icon[y*16+x];
			c = ((c&0x001f) << 10) | ((c&0x7c00) >> 10) | (c&0x03e0);
			if (visual->depth == 16)
				c = (c&0x001f) | ((c&0x7c00) << 1) | ((c&0x03e0) << 1);
			else if (visual->depth == 24 || visual->depth == 32)
				c = ((c&0x001f) << 3) | ((c&0x03e0) << 6) | ((c&0x7c00) << 9);
				
			gdk_image_put_pixel(image, x, y, c);
		}
	}

	pixmap = gdk_pixmap_new(dialog->window, 16, 16, visual->depth);

	gc = gdk_gc_new(pixmap);
	gdk_draw_image(pixmap, gc, image, 0, 0, 0, 0, 16, 16);
	gdk_gc_destroy(gc);
	gdk_image_destroy(image);
	
	return gdk_pixbuf_get_from_drawable    (NULL,
	                                        GDK_PIXMAP (pixmap),
		                                NULL,
						0,0,0,0,-1,-1);
}

static void
LoadListItems(int mcd, GtkWidget *List, GtkWidget *dialog) {
	int i;

	GtkListStore *store= gtk_list_store_new (NUM_CL,GDK_TYPE_PIXBUF, G_TYPE_STRING,
					     G_TYPE_STRING,G_TYPE_STRING,G_TYPE_STRING);
	GtkTreeIter iter;
	GdkPixbuf *pixbuf;

	for (i = 0; i < MAX_MEMCARD_BLOCKS; i++) {
		McdBlock *Info;
		gchar *state;

		Info = &Blocks[mcd-1][i];
		IconC[mcd-1][i] = 0;

		if ((Info->Flags & 0xF0) == 0xA0) {
			if ((Info->Flags & 0xF) >= 1 &&
				(Info->Flags & 0xF) <= 3) {
				state = _("Deleted");
			} else
				state = _("Free");
		} else if ((Info->Flags & 0xF0) == 0x50)
			state = _("Used");
		else
			state = _("Free");

//		if (Info->IconCount == 0) continue;

		pixbuf = SetIcon(dialog, Info->Icon, i+1);

		gtk_list_store_append (store, &iter);
		gtk_list_store_set (store, &iter,
				CL_ICON, pixbuf,
				CL_TITLE, Info->Title,
				CL_STAT, state,
				CL_NAME, Info->Name,
				CL_ID, Info->ID,
				-1);
	}

	gtk_tree_view_set_model (GTK_TREE_VIEW (List), GTK_TREE_MODEL (store));
	g_object_unref (G_OBJECT (store));
	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (List), TRUE);
	gtk_widget_show(List);
}


static void
UpdateMcdDlg(GtkWidget *widget) {
	int i;
	GladeXML *xml;
	GtkWidget *list1, *list2, *dialog;
	
	for (i = 0; i < MAX_MEMCARD_BLOCKS; i++) {
		GetMcdBlockInfo(1, i+1, &Blocks[0][i]);
		GetMcdBlockInfo(2, i+1, &Blocks[1][i]);
	}

	xml = glade_get_widget_tree (widget);
	list1 = glade_xml_get_widget (xml, "GtkCList_McdList1");
	list2 = glade_xml_get_widget (xml, "GtkCList_McdList2");
	dialog = glade_xml_get_widget (xml, "McdsDlg");
	LoadListItems(1, list1, dialog);
	LoadListItems(2, list2, dialog);
}

static void
StopTimer() {
	g_timer_stop (Gtimer);
	timer = 0;
}

void OnMcd_Clicked (GtkDialog *dialog, gint arg1, gpointer user_data)
{
	GladeXML *xml = user_data;
	

	StopTimer();

	if (arg1 == GTK_RESPONSE_CLOSE)
	{
		GtkWidget *widget;
		gchar *memcard1, *memcard2;
	
		widget = glade_xml_get_widget (xml, "GtkMcd1FSButton");
		memcard1 = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));

		widget = glade_xml_get_widget (xml, "GtkMcd2FSButton");
		memcard2 = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));

		g_object_set (pcsx_gui_prefs,
					  "memcard-memcard1", memcard1,
					  "memcard-memcard2", memcard2,
					  NULL);
			
		if ((memcard1 != NULL) && (memcard2 != NULL))
			LoadMcds (memcard1, memcard2);

		g_free (memcard1);
		g_free (memcard2);
	} 
	
	gtk_widget_destroy(GTK_WIDGET (dialog));
	
}

void on_memcard_file_changed (GtkWidget *widget, gpointer user_data) {
	gint memcard = (int) user_data;
	gchar *filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));

	if (filename != NULL) {
		LoadMcd(memcard, filename);
		UpdateMcdDlg(widget);
	}
	g_free (filename);
}

void OnMcd_Format(GtkWidget *widget, gpointer user_data) {
	GladeXML *xml;
	GtkWidget *memcard_fs;
	GtkWidget *message_dialog;
	gint result;
	gchar *str;

	gint memcard = (int) user_data;
	
	/* TODO Check if the memory card actually has data on it before displaying this */

	message_dialog = gtk_message_dialog_new (NULL, GTK_DIALOG_MODAL,
		GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
		_("Format this Memory Card?"),
		NULL);
	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (message_dialog),
		_("If you format the memory card, the card will be empty, and any existing data overwritten."),
		NULL);
	gtk_dialog_add_buttons (GTK_DIALOG (message_dialog),
		GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		_("Format card"), GTK_RESPONSE_YES,
		NULL);
	
	result = gtk_dialog_run (GTK_DIALOG (message_dialog));
	gtk_widget_destroy (message_dialog);

	if (result == GTK_RESPONSE_YES) {
		xml = glade_get_widget_tree (widget);
		if (memcard == 1)
			memcard_fs = glade_xml_get_widget (xml, "GtkMcd1FSButton");
		else
			memcard_fs = glade_xml_get_widget (xml, "GtkMcd2FSButton");
		str = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (memcard_fs));

		CreateMcd(str);
		LoadMcd(memcard, str);
		UpdateMcdDlg(widget);
		g_free (str);
	}
}

void OnMcd_Reload (GtkWidget *widget, gpointer user_data) {
	GladeXML *xml;
	GtkWidget *memcard_fs;
	gchar *str;

	gint memcard = (int) user_data;
	
	xml = glade_get_widget_tree (widget);

	if (memcard == 1) {
		memcard_fs = glade_xml_get_widget (xml, "GtkMcd1FSButton");
		str = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (memcard_fs));
	} else {
		memcard_fs = glade_xml_get_widget (xml, "GtkMcd2FSButton");
		str = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (memcard_fs));
	}

	LoadMcd (memcard, str);
	UpdateMcdDlg(widget);

	g_free (str);
}

static int copy = 0, copymcd = 0;

void OnMcd_CopyTo (GtkWidget *widget, gpointer user_data) {
	gint mcd = (gint) user_data;

	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreePath *path;
	gint *i;
	GladeXML *xml;
	GtkWidget *treeview;
	GtkTreeSelection *treesel;
	GtkWidget *btnpaste;
	gchar *str;
	unsigned char *source, *destination;

	int first_free_slot;
	
	xml = glade_get_widget_tree (widget);

	if (mcd == 1)
		treesel = gtk_tree_view_get_selection (GTK_TREE_VIEW (GtkCList_McdList1));
	else
		treesel = gtk_tree_view_get_selection (GTK_TREE_VIEW (GtkCList_McdList2));

	/* If the item selected is not reported as a 'Free' slot */
	if (gtk_tree_selection_get_selected (treesel, &model, &iter)) {
		path = gtk_tree_model_get_path (model, &iter);
		i = gtk_tree_path_get_indices (path);
		copy    = *i;
		copymcd = mcd;
		gtk_tree_path_free (path);
	}
	
	/* Determine the first free slot in the target memory card */
	first_free_slot = get_free_memcard_slot (mcd - 1);
	if (first_free_slot == -1) {
		/* No free slots available on the destination card */
		SysErrorMessage (_("No free space on memory card"),
						 _("There are no free slots available on the target memory card. Please delete a slot first."));
		return;
	}
	xml = glade_get_widget_tree (GtkCList_McdList1);

	if (mcd == 1) {
		str = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (glade_xml_get_widget (xml, "GtkMcd1FSButton")));
		source = Mcd2Data;
		copy = mcd2_row;
		destination = Mcd1Data;
		printf("Copying from card 2 to card 1 from slot %d into slot %d\n", mcd2_row, first_free_slot);
	} else {
		str = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (glade_xml_get_widget (xml, "GtkMcd2FSButton")));
		source = Mcd1Data;
		copy = mcd1_row;
		destination = Mcd2Data;
		printf("Copying from card 1 to card 2 from slot %d into slot %d\n", mcd1_row, first_free_slot);
	}
	g_return_if_fail (str != NULL);

	copy_memcard_data (source, destination, &first_free_slot, str);
	UpdateMcdDlg(widget);
}

static void
copy_memcard_data (char *from, char *to, gint *i, gchar *str) {
	memcpy(to + (*i+1) * 128, from + (copy+1) * 128, 128);
	SaveMcd((char*)str, to, (*i+1) * 128, 128);
	memcpy(to + (*i+1) * 1024 * 8, from + (copy+1) * 1024 * 8, 1024 * 8);
	SaveMcd((char*)str, to, (*i+1) * 1024 * 8, 1024 * 8);
}

static int
get_free_memcard_slot (int target_card) {
	GtkTreeModel *model;
	McdBlock *Info;
	gboolean found = FALSE;

	int i = 0;
	while (i < 15  && found == FALSE) {
		Info = &Blocks[target_card][i];
		if (g_ascii_strcasecmp (Info->Title, "") == 0) {
			//if ((Info->Flags & 0xF0) != 0x50) {
			/* i.e. State is not either Deleted or Free, not Used */
			found = TRUE;
		} else {
			i++;
		}
	}

	if (found == TRUE)
		return i;

	/* no free slots, try to find a deleted one */
	i = 0;
	while (i < 15  && found == FALSE) {
		Info = &Blocks[target_card][i];
		if ((Info->Flags & 0xF0) != 0x50) {
			found = TRUE;
		} else {
			i++;
		}
	}

	if (found == TRUE)
		return i;

	return -1;	
		

}

static void
tree_selection_changed_cb (GtkTreeSelection *selection, gpointer user_data) {
	GladeXML *xml;
	GtkWidget *tree;
	GtkTreeSelection *sel;
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreePath *path;
	
	gboolean selected;
	int i;
	unsigned char *data;

	gint memcard = (int) user_data;

	if (memcard == 1) {
		selected = gtk_tree_selection_get_selected (selection, &model, &iter);
		data = Mcd1Data;
	} else {
		selected = gtk_tree_selection_get_selected (selection, &model, &iter);
		data = Mcd2Data;
	}

	if (selected) {
		path = gtk_tree_model_get_path (model, &iter);
		i = *gtk_tree_path_get_indices (path);

		/* If a row was selected, and the row is not blank, we can now enable
		   some of the disabled widgets */
		/* Check LoadListItems for determining state */
		xml = glade_get_widget_tree (GtkCList_McdList1);
		GtkWidget *disable, *enable;
		if (memcard == 1) {
			mcd1_row = i;
			enable = glade_xml_get_widget (xml, "GtkButton_CopyTo2");
			disable = glade_xml_get_widget (xml, "GtkButton_CopyTo1");
		} else {
			mcd2_row = i;
			enable = glade_xml_get_widget (xml, "GtkButton_CopyTo1");
			disable = glade_xml_get_widget (xml, "GtkButton_CopyTo2");
		}
		gtk_widget_set_sensitive (enable, TRUE);
		gtk_widget_set_sensitive (disable, FALSE);      
	}
}

static void
on_memcard_delete (GtkWidget *widget, gpointer user_data) {
	McdBlock *Info;
	int i, xor = 0, j;
	unsigned char *data, *ptr;
	const char *str;
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreePath *path;
	gchar *filename;
	GladeXML *xml;
	gboolean selected;
	GtkWidget *tree;
	GtkTreeSelection *sel;

	gint memcard = (int) user_data;

	xml = glade_get_widget_tree (widget);
	if (memcard == 1) {
		tree = glade_xml_get_widget (xml, "GtkCList_McdList1");
		sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
		selected = gtk_tree_selection_get_selected (sel, &model, &iter);
		data = Mcd1Data;
		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (glade_xml_get_widget (xml, "GtkMcd1FSButton")));
	} else {
		tree = glade_xml_get_widget (xml, "GtkCList_McdList2");
		sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
		selected = gtk_tree_selection_get_selected (sel, &model, &iter);
		data = Mcd2Data;
		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (glade_xml_get_widget (xml, "GtkMcd2FSButton")));
	}
	printf("Will delete from memory card %s\n", filename);
	if (selected) {
		path = gtk_tree_model_get_path (model, &iter);
		i = *gtk_tree_path_get_indices (path);

		i++;
		ptr = data + i * 128;
		Info = &Blocks[memcard-1][i-1];

		if ((Info->Flags & 0xF0) == 0xA0) {
			if ((Info->Flags & 0xF) >= 1 &&
				(Info->Flags & 0xF) <= 3) { // deleted
				*ptr = 0x50 | (Info->Flags & 0xF);
			} else return;
		} else if ((Info->Flags & 0xF0) == 0x50) { // used
				*ptr = 0xA0 | (Info->Flags & 0xF);
		} else { return; }

		for (j=0; j<127; j++) xor^=*ptr++;
		*ptr = xor;

		SaveMcd((char*)filename, data, i * 128, 128);
		UpdateMcdDlg(widget);
	}
}


void UpdateMcdIcon(int mcd, GtkWidget *List) {
#if 0
	int i;
	GtkListStore *store= gtk_list_store_new (NUM_CL,GDK_TYPE_PIXBUF, G_TYPE_STRING,
					     G_TYPE_STRING,G_TYPE_STRING,G_TYPE_STRING);
	GtkTreeIter iter;
	GdkPixbuf *pixbuf;

	gtk_tree_clear_items(GTK_TREE_VIEW (List));

	for (i=0; i<MAX_MEMCARD_BLOCKS; i++) {
		McdBlock *Info;
		int *count;

		Info = &Blocks[mcd-1][i];
		count = &IconC[mcd-1][i];

		if (Info->IconCount <= 1) continue;

		(*count)++;
		if (*count == Info->IconCount) *count = 0;

	{
		gchar *state;

		Info = &Blocks[mcd-1][i];
		IconC[mcd-1][i] = 0;

		if ((Info->Flags & 0xF0) == 0xA0) {
			if ((Info->Flags & 0xF) >= 1 &&
				(Info->Flags & 0xF) <= 3) {
				state = _("Deleted");
			} else state = _("Free");
		} else if ((Info->Flags & 0xF0) == 0x50)
			state = _("Used");
		else { state = _("Free");}

//		pixbuf=SetIcon(Info->Icon, i);
		pixbuf=SetIcon(&Info->Icon[*count*16*16], i+1);
		
		gtk_list_store_append (store, &iter);
		gtk_list_store_set (store, &iter,
				CL_ICON,pixbuf,
				CL_TITLE,Info->Title,
				CL_STAT,state,
				CL_NAME,Info->Name,
				CL_ID,Info->ID,
				-1);
		g_free (state);
	}

//		SetIcon(&Info->Icon[*count*16*16], i);
	}
/*	gtk_tree_view_set_model (GTK_TREE_VIEW (List), GTK_TREE_MODEL (store));
	g_object_unref (G_OBJECT (store));
	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (List), TRUE);*/
//	gtk_widget_show(List);

#endif
}

void OnConf_Mcds() {

	GladeXML *xml;
	GtkWidget *dialog;
	GtkWidget *widget;
	GtkWidget *list1, *list2;
	GtkTreeSelection *treesel1, *treesel2;
	gchar *memcard1, *memcard2;
	gchar *str;
	GtkWidget *btn_paste;
	int i;

	xml = glade_xml_new(PACKAGE_DATA_DIR "pcsx.glade2", "McdsDlg", NULL);

	if (!xml)
	{
		g_warning("We could not load the interface!");
		return;
	}

	dialog = glade_xml_get_widget(xml, "McdsDlg");

	gtk_window_set_title(GTK_WINDOW(dialog), _("Memory Card Manager"));

	g_object_get (pcsx_gui_prefs,
				  "memcard-memcard1", &memcard1,
				  "memcard-memcard2", &memcard2,
				  NULL);
printf("Loading memcards %s and %s\n", memcard1, memcard2);	
	/*DELETE Assign default memory cards 
	if (!strlen(Config.Mcd1)) {
		str = g_strconcat (getenv("HOME"), DEFAULT_MEM_CARD_1, NULL);
		strcpy(Config.Mcd1, str);
		g_free (str);
	}

	if (!strlen(Config.Mcd2)) {
		str = g_strconcat (getenv("HOME"), DEFAULT_MEM_CARD_2, NULL);
		strcpy(Config.Mcd2, str);
		g_free (str);
	}*/

	GtkCList_McdList1 = glade_xml_get_widget(xml, "GtkCList_McdList1");
	add_columns(GTK_TREE_VIEW (GtkCList_McdList1));
	GtkCList_McdList2 = glade_xml_get_widget(xml, "GtkCList_McdList2");
	add_columns(GTK_TREE_VIEW (GtkCList_McdList2));

	treesel1 = gtk_tree_view_get_selection (GTK_TREE_VIEW (GtkCList_McdList1));
	gtk_tree_selection_set_mode (treesel1, GTK_SELECTION_SINGLE);
	g_signal_connect_data (G_OBJECT (treesel1), "changed",
						   G_CALLBACK (tree_selection_changed_cb),
						   (gpointer) 1, NULL,
						   G_CONNECT_AFTER);

	treesel2 = gtk_tree_view_get_selection (GTK_TREE_VIEW (GtkCList_McdList2));
	gtk_tree_selection_set_mode (treesel2, GTK_SELECTION_SINGLE);
	g_signal_connect_data (G_OBJECT (treesel2), "changed",
						   G_CALLBACK (tree_selection_changed_cb),
						   (gpointer) 2, NULL,
						   G_CONNECT_AFTER);
	
	for (i = 0; i < MAX_MEMCARD_BLOCKS; i++) {
		GetMcdBlockInfo (1, i+1, &Blocks[0][i]);
		GetMcdBlockInfo (2, i+1, &Blocks[1][i]);
	}

	LoadListItems(1, GtkCList_McdList1, dialog);
	LoadListItems(2, GtkCList_McdList2, dialog);

	/* Setup a handler for when Close or Cancel is clicked */
	g_signal_connect_data(GTK_OBJECT(dialog), "response",
			GTK_SIGNAL_FUNC(OnMcd_Clicked), xml, (GClosureNotify)g_object_unref, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "GtkButton_Format1");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(OnMcd_Format), (gpointer) 1, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "GtkButton_Format2");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(OnMcd_Format), (gpointer) 2, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "GtkMcd1FSButton");
	gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (widget), memcard1/*Config.Mcd1*/);
	g_signal_connect_data(GTK_OBJECT(widget), "selection-changed",
			GTK_SIGNAL_FUNC(on_memcard_file_changed), (gpointer) 1, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "GtkMcd2FSButton");
	gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (widget), memcard2/*Config.Mcd2*/);
	g_signal_connect_data(GTK_OBJECT(widget), "selection-changed",
			GTK_SIGNAL_FUNC(on_memcard_file_changed), (gpointer) 2, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "GtkButton_Reload1");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(OnMcd_Reload), (gpointer) 1, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "GtkButton_Reload2");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(OnMcd_Reload), (gpointer) 2, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "GtkButton_CopyTo1");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(OnMcd_CopyTo), (gpointer) 1, NULL, G_CONNECT_AFTER);
	gtk_widget_set_sensitive (GTK_WIDGET (widget), FALSE);

	widget = glade_xml_get_widget(xml, "GtkButton_CopyTo2");
	g_signal_connect_data(GTK_OBJECT(widget), "clicked",
			GTK_SIGNAL_FUNC(OnMcd_CopyTo), (gpointer) 2, NULL, G_CONNECT_AFTER);
	gtk_widget_set_sensitive (GTK_WIDGET (widget), FALSE);

	widget = glade_xml_get_widget(xml, "GtkButton_Delete1");
	g_signal_connect_data (GTK_OBJECT (widget), "clicked",
			GTK_SIGNAL_FUNC (on_memcard_delete), (gpointer) 1, NULL, G_CONNECT_AFTER);

	widget = glade_xml_get_widget(xml, "GtkButton_Delete2");
	g_signal_connect_data (GTK_OBJECT (widget), "clicked",
			GTK_SIGNAL_FUNC (on_memcard_delete), (gpointer) 2, NULL, G_CONNECT_AFTER);

	Gtimer = g_timer_new();
	timer = 1;

	while (gtk_events_pending()) gtk_main_iteration();

	while (timer) {
		unsigned long usecs;

		g_timer_elapsed(Gtimer, &usecs);
		if (usecs > 250000) {
			UpdateMcdIcon(1, list1);
			UpdateMcdIcon(2, list2);
			g_timer_reset(Gtimer);
		}

		while (gtk_events_pending()) gtk_main_iteration();
		usleep(10000);
	}
	// TODO Need to free the timer?
}