#include <glade/glade.h>

#include "Linux.h"
#include "Gtk2Gui.h"
#include "memcard-dlg.h"
#include "pcsx-main-win.h"

struct _PcsxMainWinPrivate {

	/* Action group to control menu and toolbar items that are dependant
	   on the plugins having been configured */
	GtkActionGroup *exec_action_group;
	
};

#define PCSX_MAIN_WIN_GET_PRIVATE(o)  (PCSX_MAIN_WIN (o)->priv)

G_DEFINE_TYPE (PcsxMainWin, pcsx_main_win, GTK_TYPE_WINDOW)

/* Function prototypes */
static void
pcsx_main_win_destroy              (GtkObject *object);

/* Boilerplate functions */
static GObject *
pcsx_main_win_constructor (GType                  type,
				guint                  n_construct_properties,
				GObjectConstructParam *construct_properties)
{
	GObject          *obj;
	PcsxMainWin *dialog;

	obj = G_OBJECT_CLASS (pcsx_main_win_parent_class)->constructor (type,
									     n_construct_properties,
									     construct_properties);

	dialog = PCSX_MAIN_WIN (obj);

	return obj;
}

static void
pcsx_main_win_class_init (PcsxMainWinClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	GtkObjectClass *gtkobject_class = GTK_OBJECT_CLASS (class);
	GtkWindowClass *gtkwindow_class = GTK_WINDOW_CLASS (class);

	gobject_class->constructor = pcsx_main_win_constructor;
/*	gobject_class->get_property = pcsx_main_win_get_property;
	gobject_class->set_property = pcsx_main_win_set_property;*/

	gtkobject_class->destroy = pcsx_main_win_destroy;

	g_type_class_add_private (class,
				  sizeof (PcsxMainWinPrivate));

	/* Signals and properties go here */

}

static void
pcsx_main_win_init (PcsxMainWin *win)
{
	PcsxMainWinPrivate *priv;

	GtkWidget *toolbar;
	GtkWidget *menubar;
	GtkWidget *vbox;
	GdkPixbuf *pixbuf;
	GtkWidget *image;
	
	GtkAccelGroup *accel_group;
	GtkTooltips *tooltips;
	
	GtkUIManager *mgr;
	GtkActionGroup *action_group;
	GError *error = NULL;
	
	priv = G_TYPE_INSTANCE_GET_PRIVATE (win,
					    PCSX_TYPE_MAIN_WIN,
					    PcsxMainWinPrivate);

	win->priv = priv;
	
	/* Initialise private variables */

	
	/* Create a vbox and add it to the window */
	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (win), vbox);

	/* Prepare the UI manager to build the menu and toolbar */
	mgr = gtk_ui_manager_new ();
	
	/* Add commands that are always enabled */
	action_group = gtk_action_group_new ("PcsxMenuActions");
	gtk_action_group_set_translation_domain (action_group, GETTEXT_PACKAGE);
	gtk_action_group_add_actions (action_group,
								  pcsx_menu_entries,
								  G_N_ELEMENTS (pcsx_menu_entries),
								  win);
	gtk_ui_manager_insert_action_group (mgr, action_group, 0);
	g_object_unref (action_group);
	
	/* Add commands that are enabled only when plugins are configured */
	action_group = gtk_action_group_new ("PcsxExecMenuActions");
	gtk_action_group_set_translation_domain (action_group, GETTEXT_PACKAGE);
	gtk_action_group_add_actions (action_group,
								  pcsx_exec_menu_entries,
								  G_N_ELEMENTS (pcsx_exec_menu_entries),
								  win);
	gtk_ui_manager_insert_action_group (mgr, action_group, 0);
	g_object_unref (action_group);

	/* Set the action group so its members can be referenced for enabling/disabling
	   if plugins are not configured */
	priv->exec_action_group = action_group;
	
	/* Load the UI from the XML file */
	gtk_ui_manager_add_ui_from_file (mgr, PACKAGE_DATA_DIR "pcsx-ui.xml", &error);
	if (error != NULL)
	{
		printf ("Error loading pcsx-ui.xml: %s\n", error->message);
		g_error_free (error);
		error = NULL;
	}
	
	/* Set up the menu */
	menubar = gtk_ui_manager_get_widget (mgr, "/MenuBar");
	gtk_box_pack_start (GTK_BOX (vbox), menubar, FALSE, FALSE, 0);	
	
	/* Set up the toolbar */
	toolbar = gtk_ui_manager_get_widget (mgr, "/ToolBar");
	gtk_box_pack_start (GTK_BOX (vbox), toolbar, FALSE, FALSE, 0);
	
	/* Enable keyboard shortcuts defined in the UI Manager */
	accel_group = gtk_accel_group_new ();
	gtk_window_add_accel_group (GTK_WINDOW (win), accel_group);     /* FIXME TODO Is this one required? */
	gtk_window_add_accel_group (GTK_WINDOW (win), gtk_ui_manager_get_accel_group (mgr));

	/* Set tooltips */
	tooltips = gtk_tooltips_new ();
	g_object_set_data (G_OBJECT (win), "tooltips", tooltips);
	
	/* Add the logo to the main area */
	pixbuf = gdk_pixbuf_new_from_file (PACKAGE_DATA_DIR "pcsx.jpg", &error);
	if (error != NULL)
	{
		printf ("Error loading pcsx.jpg: %s\n", error->message);
		g_error_free (error);
		error = NULL;
	}
	if (pixbuf) {
		image = gtk_image_new_from_pixbuf (pixbuf);
		gtk_box_pack_start (GTK_BOX (vbox), image, FALSE, FALSE, 0);
	}
	
	/* Add a logo */
	gtk_window_set_icon_from_file (GTK_WINDOW (win), PIXMAPDIR "pcsx-icon.png", &error);
	if (error != NULL)
	{
		printf ("Error loading pcsx-icon.png: %s\n", error->message);
		g_error_free (error);
		error = NULL;
	}

	/* Callbacks */
	g_signal_connect (G_OBJECT (win), "delete_event", G_CALLBACK (OnFile_Exit), NULL);
	
}

GtkWidget *
pcsx_main_win_new (GtkWindow *parent)
{
	GtkWindow *win;

	win = g_object_new (PCSX_TYPE_MAIN_WIN,
						"title", _("PCSX-df"),
						NULL);

	if (parent)
		gtk_window_set_transient_for (GTK_WINDOW (win), parent);

	return win;

}

static void
pcsx_main_win_destroy (GtkObject *object)
{
	PcsxMainWin *win;
	
printf ("Destroying PCSX main window...\n");
	win = PCSX_MAIN_WIN (object);
	
	if (win->priv->exec_action_group)
		g_object_unref (win->priv->exec_action_group);
	
	g_object_unref (win->priv);
	
/*	GTK_OBJECT_CLASS (pcsx_main_win_parent_class)->destroy (object);*/
	
printf ("Destroying PCSX main window...done\n");
}


void
pcsx_main_window_set_exec_grp_sensitive (PcsxMainWin *win, gboolean sensitive)
{
	gtk_action_group_set_sensitive (win->priv->exec_action_group, sensitive);
}
